from anylearn.sdk.artifacts.algorithm import AlgorithmArtifact
from anylearn.sdk.artifacts.artifact import Artifact, ArtifactState
from anylearn.sdk.artifacts.dataset import DatasetArtifact
from anylearn.sdk.artifacts.file import FileArtifact
from anylearn.sdk.artifacts.model import ModelArtifact
from anylearn.sdk.project import Project
from anylearn.sdk.task import Task


def get_task_output(task_id: str) -> FileArtifact:
    task = Task.from_id(task_id)
    return FileArtifact.from_id(task.output_artifact_id)



from __future__ import annotations
import time
from typing import Dict, List,Optional, Union

from anylearn.applications.algorithm_manager import sync_algorithm
from anylearn.applications.hpo_experiment import HpoExperiment
from anylearn.applications.quickstart import (
    _get_or_create_dataset,
    _upload_dataset,
)
from anylearn.applications.utils import generate_random_name
from anylearn.interfaces import Project
from anylearn.interfaces.resource import ResourceUploader
from anylearn.storage.db import DB
from anylearn.utils.errors import AnyLearnException


def run_hpo(hpo_search_space: dict,
            hpo_max_runs: int=10,
            hpo_max_duration: str="24h",
            hpo_tuner_name: str="TPE",
            hpo_mode: str="maximize",
            hpo_concurrency: int=1,
            project_name: Optional[str]=None,
            algorithm_id: Optional[str]=None,
            algorithm_name: Optional[str]=None,
            algorithm_dir: Optional[str]=None,
            algorithm_archive: Optional[str]=None,
            algorithm_entrypoint: Optional[str]=None,
            algorithm_output: Optional[str]=None,
            mirror_name: Optional[str]="QUICKSTART",
            dataset_id: Optional[str]=None,
            dataset_dir: Optional[str]=None,
            dataset_archive: Optional[str]=None,
            resource_uploader: Optional[ResourceUploader]=None,
            resource_polling: Union[float, int]=5,
            resource_timeout: Union[float, int]=120,
            dataset_hyperparam_name: str="dataset",
            resource_request: Optional[List[Dict[str, Dict[str, int]]]]=None) -> HpoExperiment:
    """
    超参数自动调优接口。

    仅需提供调优配置参数和本地资源相关信息，
    即可在本地启动针对自定义算法/数据集的超参数自动调优，
    调优过程中使用Anylearn后端引擎进行多次训练，
    汇总到同一个训练项目中。

    目前接受的调优配置包括（详情参见本接口参数列表）：

    - 超参数搜索空间
    - 最大调优任务数（到量终止调优）
    - 最长调优总时间（到时终止调优）
    - 调优算法
    - 调优模式（最大化或最小化指标）
    - 端口绑定

    与本地快速训练类似，
    本接口封装了Anylearn从零启动训练的一些列流程。
    参见 :func:`~anylearn.applications.quickstart.quick_train` 。
    此外，自动调优的配置和构建过程也封装在接口内。

    Parameters
    ----------
    hpo_search_space : :obj:`dict`
        超参数搜索空间，格式详询 :obj:`NNI` 文档。
    hpo_max_runs : :obj:`int`, optional
        最大调优任务数，达成即终止实验。默认为 :obj:`10` 。
    hpo_max_duration : :obj:`str`, optional
        最长调优总时间，达成即终止实验。
        格式为 :obj:`数字 + s|m|h|d`，
        例如： 半天 :obj:`0.5d` ， 十分钟 :obj:`10m` 。
        默认为 :obj:`24h` 。
    hpo_tuner_name : :obj:`str`, optional
        调优算法名称。
        详见 `NNI内置调优算法 <https://nni.readthedocs.io/zh/stable/builtin_tuner.html>`_ 。
        默认为 :obj:`TPE` 。
    hpo_mode : :obj:`str`, optional
        调优模式，即，最大化或最小化指标。
        可选项： :obj:`maximize` 或 :obj:`minimize` 。
        默认为 :obj:`maximize` 。
    hpo_concurrency : :obj:`int`, optional
        调优期望并行度，即期望同时运行的任务数，实际并行度视后端资源调度情况而定。
        默认为 :obj:`1` 。
    project_name : :obj:`str`, optional
        将要创建的训练项目名称。如已传远程 :obj:`project_id` 则忽略。
    algorithm_id : :obj:`str`, optional
        已在Anylearn远程注册的算法ID。
    algorithm_name: :obj:`str`, optional
        指定的算法名称。
        注：同一用户的自定义算法的名称不可重复。
        如有重复，则复用已存在的同名算法，
        算法文件将被覆盖并提升版本。
        原有版本仍可追溯。
    algorithm_dir : :obj:`str`, optional
        本地算法目录路径。
    algorithm_archive : :obj:`str`, optional
        本地算法压缩包路径。
    algorithm_entrypoint : :obj:`str`, optional
        启动训练的入口命令。
    algorithm_output : :obj:`str`, optional
        训练输出模型的相对路径（相对于算法目录）。
    mirror_name : :obj:`str`, optional
        快速训练镜像名，默认为QUICKSTART。
    dataset_id : :obj:`str`, optional
        已在Anylearn远程注册的数据集ID。
    dataset_dir : :obj:`str`, optional
        本地数据集目录路径。
    dataset_archive : :obj:`str`, optional
        本地数据集压缩包路径。
    resource_uploader : :obj:`ResourceUploader`, optional
        资源上传实现。
        默认使用系统内置的同步上传器 :obj:`SyncResourceUploader` 。
    resource_polling : :obj:`float|int`, optional
        资源上传中轮询资源状态的时间间隔（单位：秒）。
        默认为5秒。
    resource_timeout : :obj:`float|int`, optional
        轮询状态的超时时长（单位：秒）。
        默认为120秒。
    dataset_hyperparam_name : :obj:`str`, optional
        启动训练时，数据集路径作为启动命令参数传入算法的参数名。
        需指定长参数名，如 :obj:`--data` ，并省略 :obj:`--` 部分传入。
        数据集路径由Anylearn后端引擎管理。
        默认为 :obj:`dataset` 。
    hyperparams : :obj:`dict`, optional
        训练超参数字典。
        超参数将作为训练启动命令的参数传入算法。
        超参数字典中的键应为长参数名，如 :obj:`--param` ，并省略 :obj:`--` 部分传入。
        默认为空字典。
    resource_request : :obj:`List[Dict[str, Dict[str, int]]]`, optional
        单次训练所需计算资源的请求。
        如未填，则使用Anylearn后端的:obj:`default`资源组中的默认资源套餐。

    Returns
    -------
    HpoExperiment
        调参实验对象
    """
    # Algorithm
    algo = sync_algorithm(
        id=algorithm_id,
        name=algorithm_name,
        dir_path=algorithm_dir,
        archive_path=algorithm_archive,
        mirror_name=mirror_name,
        uploader=resource_uploader,
        polling=resource_polling,
    )

    # Dataset
    dset, dataset_archive, dataset_checksum = _get_or_create_dataset(
        id=dataset_id,
        dir_path=dataset_dir,
        archive_path=dataset_archive
    )
    if dataset_archive and dataset_checksum:
        # Local dataset registration
        _upload_dataset(dataset=dset,
                        dataset_archive=dataset_archive,
                        uploader=resource_uploader,
                        polling=resource_polling)
        DB().create_local_dataset(id=dset.id, checksum=dataset_checksum)

    # Project
    if not project_name:
        project_name = f"PROJ_{generate_random_name()}"
    project = Project(name=project_name,
                      description="SDK_HPO_EXPERIMENT",
                      datasets=[dset.id])
    project.save()

    hpo = HpoExperiment(
        project_id=project.id,
        algorithm_id=algo.id,
        dataset_id=dset.id,
        hpo_search_space=hpo_search_space,
        hpo_max_runs=hpo_max_runs,
        hpo_max_duration=hpo_max_duration,
        hpo_tuner_name=hpo_tuner_name,
        hpo_mode=hpo_mode,
        hpo_concurrency=hpo_concurrency,
        dataset_hyperparam_name=dataset_hyperparam_name,
        resource_request=resource_request,
    )
    hpo.save()

    st = time.time()
    while not hpo.hpo_id:
        time.sleep(resource_polling)
        if time.time() > st + resource_timeout:
            raise AnyLearnException("调参实验启动超时")
        hpo.get_detail()

    return hpo



import os
from pathlib import Path
from typing import Optional, Dict, List


__DEFAULT_HOST__ = "https://anylearn.nelbds.cn"


ARTIFACT_IDS = 'ANYLEARN_ARTIFACT_IDS'
TASK_ID = 'ANYLEARN_TASK_ID'
HOST = 'ANYLEARN_HOST'
TOKEN = 'ANYLEARN_AUTH_TOKEN'
REFRESH_TOKEN = 'ANYLEARN_AUTH_REFRESH_TOKEN'


def get_artifact_ids() -> List[str]:
    ids_str = os.environ.get(ARTIFACT_IDS, None)
    if not ids_str:
        return []
    return ids_str.replace(" ", "").split(",")


def get_artifact_paths() -> List[Path]:
    ids = get_artifact_ids()
    paths = {}
    for id_ in ids:
        p = os.environ.get(id_, None)
        if not p:
            continue
        p = Path(p)
        if not p.exists():
            continue
        paths[id_] = p
    return paths


def get_task_id(default: Optional[str]=None):
    return os.environ.get(TASK_ID, default)


def get_auth() -> Dict[str, Optional[str]]:
    return {
        'host': os.environ.get(HOST, __DEFAULT_HOST__),
        'token': os.environ.get(TOKEN, None),
        'refresh_token': os.environ.get(REFRESH_TOKEN, None),
    }


def set_token(token: str) -> None:
    os.environ[TOKEN] = token



import anylearn.env as env
from anylearn.applications import (
    report_intermediate_metric,
    report_final_metric,
    run,
    sync_algorithm,
    quick_train,
    upload,
    upload_and_run,
)
from anylearn.config import (
    AnyLearnAuthException,
    AnylearnConfig,
    init_sdk,
)
from anylearn.sdk import (
    DatasetArtifact,
    FileArtifact,
    ModelArtifact,
    Task,
    get_task_output,
)


def get_dataset(full_name: str) -> DatasetArtifact:
    return DatasetArtifact.from_full_name(full_name)


def get_model(full_name: str) -> ModelArtifact:
    return ModelArtifact.from_full_name(full_name)


# Detect auth info in env and init config
auth = env.get_auth()
if auth['host'] and auth['token']:
    AnylearnConfig.init_cluster_by_token(**auth)
    AnylearnConfig.init_workspace()
    AnylearnConfig.check_git()
    # Token may have expired then been refreshed
    env.set_token(AnylearnConfig.token)



from git import RemoteProgress
from tqdm import tqdm

from ..utils import logger


class GitProgressPrinter(RemoteProgress):
    """
    Handler implementing GitPython's RemoteProgress delegate
    so as to print git operation progress.
    """

    """Reformed possible operations"""
    OPS = {
        RemoteProgress.COUNTING: "COUNTING",
        RemoteProgress.COMPRESSING: "COMPRESSING",
        RemoteProgress.WRITING: "WRITING",
        RemoteProgress.RECEIVING: "RECEIVING",
        RemoteProgress.RESOLVING: "RESOLVING",
        RemoteProgress.FINDING_SOURCES: "FINDING SOURCES",
        RemoteProgress.CHECKING_OUT: "CHECKING OUT",
    }

    """Progress bar object (tqdm instance)"""
    pbar = None

    def update(self, op_code, cur_count, max_count=None, message=''):
        """
        Delegate method to handle git progress update
        """
        # Determine the right operation
        op = -1
        for _op in self.OPS.keys():
            if op_code in [
                _op,
                _op + RemoteProgress.BEGIN,
                _op + RemoteProgress.END,
            ]:
                op = _op
                break
        if op == -1:
            logger.warning("Broken progress")

        # New operation: init tqdm
        if op_code == op + RemoteProgress.BEGIN:
            self.pbar = tqdm(total=(int(max_count) or 100),
                             initial=(int(cur_count) or 0),
                             desc=self.OPS[op])
        # End of operation: close tqdm
        elif op_code == op + RemoteProgress.END and self.pbar is not None:
            self.pbar.update(int(cur_count) - self.pbar.n)
            self.pbar.close()
        # Progressing: update tqdm
        elif self.pbar is not None:
            self.pbar.update(int(cur_count) - self.pbar.n)
        else:
            logger.warning("Broken progress")



import click

import anylearn.cli.add_algorithm
import anylearn.cli.add_dataset


@click.group("add")
def commands():
    """
    Add local algorithm or dataset to local Anylearn project.
    """
    pass

commands.add_command(anylearn.cli.add_algorithm.commands)
commands.add_command(anylearn.cli.add_dataset.commands)



from __future__ import annotations

from anylearn.sdk.artifacts.artifact import Artifact
from anylearn.utils.api import get_with_token, url_base
from anylearn.utils.errors import AnyLearnException


class AlgorithmArtifact(Artifact):
    @classmethod
    def from_full_name(cls, full_name: str) -> AlgorithmArtifact:
        res = get_with_token(
            f"{url_base()}/algorithm/query",
            params={'fullname': full_name},
        )
        if not res or not isinstance(res, list):
            raise AnyLearnException("Request failed")
        return AlgorithmArtifact(**res[0])



import click
from requests.models import HTTPError

from anylearn.cli.utils import (
    check_config,
    check_connection,
    cmd_error,
    cmd_info,
    cmd_success,
    get_cmd_command,
)
from anylearn.interfaces.train_task import TrainTask


_option_async_download = click.option(
    '--async-download',
    is_flag=True,
    default=False,
    help="Download in asynchronous mode."
)


@click.group("download")
def commands():
    """
    Download remote task results of training.
    """
    pass


@commands.command()
@_option_async_download
@check_config()
@check_connection()
@get_cmd_command()
@click.argument('task_id')
@click.option(
    '-s', '--save-path',
    prompt=True,
    help="Save path of task result download (absolute path)."
)
def training(task_id: str, save_path: str, async_download: bool=False):
    """
    Get training task's results.
    """
    try:
        train_task = TrainTask(id=task_id, load_detail=True)
        cmd_info(msg="DOWNLOADING...")
        res = train_task.download_results(save_path=save_path, async_download=async_download)
    except Exception as e:
        cmd_error(msg=f"{e}")
        cmd_info(msg="STOP DOWNLOAD")
        return
    cmd_success(msg=f"DOWNLOAD! {save_path}\\{res}")



from __future__ import annotations

import cgi
import os
from pathlib import Path
from typing import Optional, Union

import requests

from anylearn.config import AnylearnConfig
from anylearn.utils.api import get_with_token, post_with_token, url_base
from anylearn.utils.errors import AnyLearnException
from anylearn.utils.util import ensure_dir


class Compression:
    def __init__(self, *_, **kwargs):
        self.id = kwargs.pop('id', None)
        self.parent_id = kwargs.pop('parent_id', None)
        self.sub_path = kwargs.pop('subpath', None)
        self.filename = kwargs.pop('filename', None)
        self.state = kwargs.pop('state', None)
        self.creator_id = kwargs.pop('creator_id', None)
        self.created_at = kwargs.pop('created_at', None)

    @classmethod
    def from_id_and_artifact_id(
        cls,
        id_: str,
        artifact_id: str
    ) -> Compression:
        res = get_with_token(
            f"{url_base()}/resource/compression",
            params={
                'file_id': artifact_id,
                'compression_id': id_,
            },
        )
        if not res or not isinstance(res, list):
            raise AnyLearnException("Request failed")
        return Compression(**res[0])

    @classmethod
    def create(cls, artifact_id: str, sub_path: str) -> Compression:
        res = post_with_token(
            f"{url_base()}/resource/compression",
            data={
                'file_id': artifact_id,
                'subpath': sub_path,
            },
        )
        if not res or 'data' not in res:
            raise AnyLearnException("Request failed")
        compression_id = res['data']
        return Compression.from_id_and_artifact_id(
            id_=compression_id,
            artifact_id=artifact_id,
        )

    def ready_to_download(self) -> bool:
        return self.state == CompressionState.FINISHED

    def on_error(self) -> bool:
        return self.state in [
            CompressionState.FAILED,
            CompressionState.INVALID,
        ]

    def reload(self) -> None:
        res = get_with_token(
            f"{url_base()}/resource/compression",
            params={
                'file_id': self.parent_id,
                'compression_id': self.id,
            },
        )
        if not res or not isinstance(res, list):
            raise AnyLearnException("Request failed")
        self.__init__(**res[0])

    def download(
        self,
        dest_dir: Optional[Union[os.PathLike, str]]=None,
    ) -> Path:
        assert self.ready_to_download()

        root_dir = Path(dest_dir or AnylearnConfig.workspace_path / "tmp")
        ensure_dir(root_dir)

        headers = {'Authorization': f"Bearer {AnylearnConfig.token}"}
        res = requests.get(
            url=f"{url_base()}/resource/download",
            headers=headers,
            params={
                'file_id': self.parent_id,
                'compression_id': self.id,
                'token': AnylearnConfig.token,
            },
        )
        res.raise_for_status()

        content_header = res.headers.get('Content-Disposition')
        if content_header:
            _, params = cgi.parse_header(content_header)
            filename = params['filename']
            with open(root_dir / filename, "wb") as f:
                f.write(res.content)
            return root_dir / filename
        else:
            raise AnyLearnException(f"Failed to download {self.id}")


class CompressionState:
    CREATED = 0
    IN_PROGRESS = 1
    FINISHED = 2
    FAILED = -1
    INVALID = -2



from pathlib import Path
import responses
from unittest import TestCase

from anylearn.config import AnylearnConfig
from anylearn.storage.db import DB
from .config import (
    test_host,
    test_username,
    test_password,
    test_workspace
)

class BaseTestCase(TestCase):
    @responses.activate
    def setUp(self):
        responses.add(responses.POST, url=self._url("user/login"),
                      json={
                          'id': "USER001",
                          'refresh_token': "TEST_REFRESH_TOKEN",
                          'token': "TEST_TOKEN",
                          'username': test_username
                      },
                      status=200)
        AnylearnConfig.init(
            cluster_address=test_host,
            username=test_username,
            password=test_password,
            workspace=test_workspace,
            disable_git=True,
        )
        DB().force_init()
    
    def tearDown(self):
        AnylearnConfig.clear_workspace()

    def _url(self, route):
        return f"{test_host}/api/{route}"



from pathlib import Path

from anylearn.analyzers import ArtifactCollector
from tests.base_test_case import BaseTestCase


class TestArtifactCollector(BaseTestCase):
    def test_artifact_collection_cases(self):
        file_path = Path(__file__).parent.parent / \
            "assets" / \
            "analyzers" / \
            "artifact_collection_cases.py"
        names = ArtifactCollector().collect(file_path)
        self.assertIn('datasets', names)
        self.assertIn('models', names)
        self.assertIn('task_outputs', names)
        self.assertEqual(
            names['datasets'],
            set([
                'my/dataset1',
                'my/dataset2',
                'my/dataset3',
                'my/dataset4',
                'my/dataset5',
            ]),
        )
        self.assertEqual(
            names['models'],
            set([
                'my/model1',
                'my/model2',
            ]),
        )
        self.assertEqual(
            names['task_outputs'],
            set([
                'mytask1',
                'mytask2',
            ]),
        )



from datetime import datetime, timedelta
from typing import Dict, List
import uuid

def utc_plus_8():
    return datetime.utcnow() + timedelta(hours=8)


def generate_primary_key(type_name):
    """用于生成ID，给一个四个字符的类别返回一个32位长的ID

    :param type_name: 4位类别名
    :returns: 32位ID

    """
    return type_name + ''.join(uuid.uuid1().__str__().split('-'))[4:]


def logs_beautify(logs: List[Dict], debug: bool=False):
    tz = datetime.now().astimezone().tzinfo
    logs = [
        f"[{str(datetime.fromtimestamp(int(l['offset']) / 1000, tz=tz))}] {l['text']}"
        for l in logs
        if not l['text'].startswith("[Anylearn]") or debug
    ]
    return logs



import copy
import json
import time
from pathlib import Path
from requests import HTTPError
from threading import Thread
from typing import Dict, List, Optional, Tuple, Union

from anylearn.sdk import (
    AlgorithmArtifact,
    Artifact,
    DatasetArtifact,
    ModelArtifact,
    Project as SDKProject,
)
from anylearn.utils.cmd import cmd_error, cmd_warning
from anylearn.utils.func import generate_primary_key
from anylearn.applications.train_profile import TrainProfile
from .algorithm_manager import sync_algorithm
from .utils import (
    _check_resource_input,
    _get_archive_checksum,
    _get_or_create_resource_archive,
    generate_random_name,
    get_mirror_by_name,
    make_name_by_path,
)
from ..interfaces import (
    Project,
    QuotaGroup,
    TrainTask,
)
from ..interfaces.resource import (
    Algorithm,
    Dataset,
    Model,
    Resource,
    ResourceState,
    ResourceUploader,
    SyncResourceUploader,
)
from ..storage.db import DB
from ..utils import logger
from ..utils.errors import (
    AnyLearnException,
    AnyLearnMissingParamException,
    AnylearnRequiredLocalCommitException,
)


def _get_or_create_dataset(id: Optional[str]=None,
                           dir_path: Optional[Union[str, Path]]=None,
                           archive_path: Optional[str]=None):
    if not any([id, dir_path, archive_path]):
        return None, None, None
    try:
        dset = Dataset(id=id, load_detail=True)
        return dset, None, None
    except:
        if not any([dir_path, archive_path]):
            raise AnyLearnMissingParamException((
                "ID provided does not exist and none of "
                "['dir_path', 'archive_path'] "
                "is specified."
            ))
        name = make_name_by_path(dir_path or archive_path)
        archive_path = _get_or_create_resource_archive(
            name=name,
            dir_path=dir_path,
            archive_path=archive_path
        )
        checksum = _get_archive_checksum(archive_path)
        local_id = DB().find_local_dataset_by_checksum(checksum=checksum)
        if local_id:
            try:
                return Dataset(id=local_id, load_detail=True), None, None
            except:
                logger.warning(
                    f"Local dataset ({local_id}) "
                    "has been deleted remotely, "
                    "forced to re-registering dataset."
                )
                DB().delete_local_dataset(id=local_id)
        dset = Dataset(name=name, description="SDK_QUICKSTART",
                       public=False,
                       filename=f"{name}.zip",
                       is_zipfile=True)
        dset.save()
        return dset, archive_path, checksum


def _get_or_create_model(id: Optional[str]=None,
                         dir_path: Optional[Union[str,Path]]=None,
                         archive_path: Optional[str]=None,
                         algorithm: Optional[Algorithm]=None):
    _check_resource_input(id, dir_path, archive_path)
    try:
        model = Model(id=id, load_detail=True)
        return model, None, None
    except:
        if not any([dir_path, archive_path]):
            raise AnyLearnMissingParamException((
                "ID provided does not exist and none of "
                "['dir_path', 'archive_path'] "
                "is specified."
            ))
        if not algorithm or not algorithm.id:
            raise AnyLearnMissingParamException(
                "Parameter 'algorithm' should be specified "
                "when using local models."
            )
        name = make_name_by_path(dir_path or archive_path)
        archive_path = _get_or_create_resource_archive(
            name=name,
            dir_path=dir_path,
            archive_path=archive_path
        )
        checksum = _get_archive_checksum(archive_path)
        local_id = DB().find_local_model_by_checksum(checksum=checksum)
        if local_id:
            try:
                # Fetch remote model and update (eventually) related algo
                model = Model(id=local_id, load_detail=True)
                model.algorithm_id = algorithm.id
                model.save()
                return model, None, None
            except:
                logger.warning(
                    f"Local model ({local_id}) "
                    "has been deleted remotely, "
                    "forced to re-registering model."
                )
                DB().delete_local_model(id=local_id)
        model = Model(name=name, description="SDK_QUICKSTART",
                      public=False,
                      filename=f"{name}.zip",
                      is_zipfile=True,
                      algorithm_id=algorithm.id)
        model.save()
        return model, archive_path, checksum


def _upload_dataset(dataset: Dataset,
                    dataset_archive: str,
                    uploader: Optional[ResourceUploader]=None,
                    polling: Union[float, int]=5):
    if not uploader:
        uploader = SyncResourceUploader()
    t_dataset = Thread(target=Resource.upload_file,
                    kwargs={
                        'resource_id': dataset.id,
                        'file_path': dataset_archive,
                        'uploader': uploader,
                    })
    logger.info(f"Uploading dataset {dataset.name}...")
    t_dataset.start()
    t_dataset.join()
    finished = [ResourceState.ERROR, ResourceState.READY]
    while dataset.state not in finished:
        time.sleep(polling)
        dataset.get_detail()
    if dataset.state == ResourceState.ERROR:
        raise AnyLearnException("Error occured when uploading dataset")
    logger.info("Successfully uploaded dataset")


def _upload_model(model: Model,
                  model_archive: str,
                  uploader: Optional[ResourceUploader]=None,
                  polling: Union[float, int]=5):
    if not uploader:
        uploader = SyncResourceUploader()
    t_model = Thread(target=Resource.upload_file,
                    kwargs={
                        'resource_id': model.id,
                        'file_path': model_archive,
                        'uploader': uploader,
                    })
    logger.info(f"Uploading dataset {model.name}...")
    t_model.start()
    t_model.join()
    finished = [ResourceState.ERROR, ResourceState.READY]
    while model.state not in finished:
        time.sleep(polling)
        model.get_detail()
    if model.state == ResourceState.ERROR:
        raise AnyLearnException("Error occured when uploading model")
    logger.info("Successfully uploaded model")


def _get_or_create_default_project():
    try:
        return Project.get_my_default_project()
    except:
        return Project.create_my_default_project()


def _get_or_create_project(project_id: Optional[str]=None,
                           project_name: Optional[str]=None):
    try:
        return Project(id=project_id, load_detail=True)
    except:
        name = project_name or f"PROJ_{generate_random_name()}"
        description = project_name or "SDK_QUICKSTART"
        project = Project(name=name, description=description)
        project.save()
        return project


def _format_resource_request(resource_request: List[Dict[str, Dict[str, int]]]):
    if not resource_request:
        return None
    # Resource group by ID or name
    for i, req in enumerate(copy.deepcopy(resource_request)):
        for key, val in req.items():
            if key in ['default', 'besteffort'] or key.startswith('QGRP'):
                # Already ID
                continue
            try:
                qid = QuotaGroup(name=key, load_detail=True).id
            except AnyLearnException:
                raise AnyLearnException(f"Failed to find QuotaGroup: {key}")
            resource_request[i][qid] = val
            del resource_request[i][key]
    return resource_request


def _create_train_task(name: str,
                       algorithm: Algorithm,
                       project: Project,
                       hyperparams: dict,
                       hyperparams_prefix: str="--",
                       hyperparams_delimeter: str=" ",
                       mounts: Optional[Union[Dict[str, List[Resource]], Dict[str, Resource]]]=None,
                       artifacts: Optional[List[Artifact]]=None,
                       algorithm_git_ref: Optional[str]=None,
                       resource_request: Optional[List[Dict[str, Dict[str, int]]]]=None,
                       description: Optional[str]=None,			
                       entrypoint: Optional[str]=None,
                       output: Optional[str]=None,
                       algorithm_dir: Optional[Union[str, Path]]=None,
                       mirror_name: Optional[str]="QUICKSTART",
                       num_nodes: Optional[int]=1,
                       nproc_per_node: Optional[int]=1):
    resource_ids = set()
    train_params = hyperparams
    for k, v in mounts.items():
        if isinstance(v, list):
            resource_ids.update([v_item.id for v_item in v])
            train_params[k] = [f"\"${v_item.id}\"" for v_item in v]
        else:
            resource_ids.add(v.id)
            train_params[k] = f"\"${v.id}\""
    for artifact in artifacts:
        resource_ids.add(artifact.id)
    train_task = TrainTask(
        name=name,
        project_id=project.id,
        algorithm_id=algorithm.id,
        algorithm_git_ref=algorithm_git_ref,
        files=list(resource_ids),
        train_params=json.dumps(train_params),
        train_params_prefix=hyperparams_prefix,
        train_params_delimeter=hyperparams_delimeter,
        resource_request=resource_request,
        description=description,
        num_nodes=num_nodes,
        nproc_per_node=nproc_per_node,
        mirror_id=get_mirror_by_name(mirror_name).id,
    )
    if entrypoint is not None:
        train_task.entrypoint = entrypoint
    if output is not None:
        output_path = Path(output)
        output_path_ok = (
            not output_path.is_absolute()
            and '..' not in output_path.parts
        )
        if output_path_ok and algorithm_dir is not None:
            output_path_joined = algorithm_dir / output_path
            output_path_ok &= (
                not output_path_joined.exists()
                or (
                    not output_path_joined.is_symlink()
                    and output_path_joined.is_dir()
                    and len([*output_path_joined.iterdir()]) == 0
                )
            )
        if not output_path_ok:
            raise AnyLearnException(
                f'Invalid output path. A relative path without ".." required, '
                f'and that path must be pointing at nothing or an empty '
                f'directory (symlink not allowed). Got '
                f'"{output_path_ok}".'
            )
        train_task.output = output
    train_task.save()
    train_task.get_detail()
    return train_task


def upload(
    algorithm_cloud_name: str,
    algorithm_entrypoint: str=None,
    algorithm_output: Optional[str]=None,
    algorithm_local_dir: Optional[Union[str, Path]]='.',
    algorithm_force_update: bool=False,
) -> Tuple[Algorithm, str]:
    try:
        return sync_algorithm(
            name=algorithm_cloud_name,
            dir_path=algorithm_local_dir,
            force=algorithm_force_update,
        )
    except AnylearnRequiredLocalCommitException:
        # Notify possible usage of algorithm_force_update=True
        raise AnylearnRequiredLocalCommitException(
            "Local algorithm code has uncommitted changes. "
            "Please commit your changes or "
            "specify `algorithm_force_update=True` "
            "to let Anylearn make an auto-commit."
    )


def run(
    algorithm_cloud_name: str,
    algorithm_entrypoint: str,
    algorithm_output: Optional[str]='./output/',
    algorithm_hyperparams: dict={},
    dataset_cloud_names: Optional[List[str]]=None,
    model_cloud_names: Optional[List[str]]=None,
    project_name: Optional[str]=None,
    image_name: Optional[str]="QUICKSTART",
    quota_group_request: Optional[Dict[str, int]]=None,
) -> Optional[TrainTask]:
    kw = locals()

    # Fetch algorithm by full name
    try:
        algo = AlgorithmArtifact.from_full_name(algorithm_cloud_name)
    except HTTPError as e:
        status_code = e.response.status_code
        if status_code == 404:
            cmd_error(f"Algorithm {algorithm_cloud_name} not found.")
        else:
            cmd_error(
                f"Error fetching algorithm {algorithm_cloud_name}, "
                "status code {status_code}."
            )
        return None
    kw['algorithm_id'] = algo.id

    # Fetch project by full name
    if project_name:
        try:
            project = SDKProject.from_full_name(kw.pop('project_name'))
        except HTTPError as e:
            status_code = e.response.status_code
            if status_code == 404:
                cmd_error(f"Project {project_name} not found.")
            else:
                cmd_error(
                    f"Error fetching project {project_name}, "
                    "status code {status_code}."
                )
            return None
        kw['project_id'] = project.id

    t, _, _, _ = quick_train(**kw)
    return t


def upload_and_run(
    algorithm_cloud_name: str,
    algorithm_entrypoint: str,
    algorithm_output: Optional[str]='./output/',
    algorithm_hyperparams: dict={},
    algorithm_local_dir: Optional[Union[str, Path]]='.',
    algorithm_force_update: bool=False,
    dataset_cloud_names: Optional[List[str]]=None,
    model_cloud_names: Optional[List[str]]=None,
    project_name: Optional[str]=None,
    image_name: Optional[str]="QUICKSTART",
    quota_group_request: Optional[Dict[str, int]]=None,
) -> TrainTask:
    kw = locals()

    # Fetch project by full name
    if project_name:
        try:
            project = SDKProject.from_full_name(kw.pop('project_name'))
        except HTTPError as e:
            status_code = e.response.status_code
            if status_code == 404:
                cmd_error(f"Project {project_name} not found.")
            else:
                cmd_error(
                    f"Error fetching project {project_name}, "
                    "status code {status_code}."
                )
            return None
        kw['project_id'] = project.id

    t, _, _, _ = quick_train(**kw)
    return t


def quick_train(algorithm_id: Optional[str]=None,
                algorithm_name: Optional[str]=None,
                algorithm_cloud_name: Optional[str]=None,
                algorithm_dir: Optional[Union[str, Path]]=None,
                algorithm_local_dir: Optional[Union[str, Path]]=None,
                algorithm_force_update: bool=False,
                algorithm_git_ref: Optional[str]=None,
                dataset_hyperparam_name: str="dataset",
                dataset_id: Optional[Union[List[str], str]]=None,
                dataset_dir: Optional[Union[str, Path]]=None,
                dataset_archive: Optional[str]=None,
                dataset_cloud_names: Optional[List[str]]=None,
                model_hyperparam_name: str="model",
                model_id: Optional[Union[List[str], str]]=None,
                model_cloud_names: Optional[List[str]]=None,
                pretrain_hyperparam_name: str="pretrain",
                pretrain_task_id: Optional[Union[List[str], str]]=None,
                project_id: Optional[str]=None,
                project_name: Optional[str]=None,
                entrypoint: Optional[str]=None,
                algorithm_entrypoint: Optional[str]=None,
                output: Optional[str]=None,
                algorithm_output: Optional[str]='./output/',
                mirror_name: Optional[str]="QUICKSTART",
                image_name: Optional[str]="QUICKSTART",
                resource_uploader: Optional[ResourceUploader]=None,
                resource_polling: Union[float, int]=5,
                hyperparams: dict={},
                hyperparams_prefix: str="--",
                hyperparams_delimeter: str=" ",
                algorithm_hyperparams: dict={},
                algorithm_hyperparams_prefix: str="--",
                algorithm_hyperparams_delimeter: str=" ",
                resource_request: Optional[List[Dict[str, Dict[str, int]]]]=None,
                quota_group_name: Optional[str]=None,
                quota_group_request: Optional[Dict[str, int]]=None,
                task_description: Optional[str]=None,
                num_nodes: Optional[int]=1,
                nproc_per_node: Optional[int]=1):
    """
    本地算法快速训练接口。

    仅需提供本地资源和训练相关的信息，
    即可在Anylearn后端引擎启动自定义算法/数据集的训练：
    
    - 算法路径（文件目录或压缩包）
    - 数据集路径（文件目录或压缩包）
    - 训练启动命令
    - 训练输出路径
    - 训练超参数

    本接口封装了Anylearn从零启动训练的一系列流程：

    - 算法注册、上传
    - 数据集注册、上传
    - 训练项目创建
    - 训练任务创建

    本地资源初次在Anylearn注册和上传时，
    会在本地记录资源的校验信息。
    下一次调用快速训练或快速验证接口时，
    如果提供了相同的资源信息，
    则不再重复注册和上传资源，
    自动复用远程资源。

    如有需要，也可向本接口传入已在Anylearn远程注册的算法或数据集的ID，
    省略资源创建的过程。

    Parameters
    ----------
    algorithm_id : :obj:`str`, optional
        已在Anylearn远程注册的算法ID。
    algorithm_name: :obj:`str`, optional
        指定的算法名称。
        注：同一用户的自定义算法的名称不可重复。
        如有重复，则复用已存在的同名算法，
        算法文件将被覆盖并提升版本。
        原有版本仍可追溯。
        自0.17.0版本起，此参数被标记为废弃，将于0.18.0版本中移除。
        请使用 :obj:`algorithm_cloud_name` 作为替代。
        
        .. deprecated:: 0.17.0
            
            use :obj:`algorithm_cloud_name` instead.
            remove in 0.18.0.
    algorithm_cloud_name: :obj:`str`, optional
        指定的算法在Anylearn云环境中的名称。
        同一用户的自定义算法的名称不可重复。
        如有重复，则复用已存在的同名算法，
        算法文件将被覆盖并提升版本。
        原有版本仍可追溯。
    algorithm_dir : :obj:`str`, optional
        本地算法目录路径。
        自0.17.0版本起，此参数被标记为废弃，将于0.18.0版本中移除。
        请使用 :obj:`algorithm_local_dir` 作为替代。
        
        .. deprecated:: 0.17.0
            
            use :obj:`algorithm_local_dir` instead.
            remove in 0.18.0.
    algorithm_local_dir : :obj:`str`, optional
        本地算法目录路径。
    algorithm_git_ref : :obj:`str`, optional
        算法Gitea代码仓库的版本号（可以是commit号、分支名、tag名）。
        使用本地算法时，如未提供此参数，则取本地算法当前分支名。
    algorithm_force_update : :obj:`bool`, optional
        在同步算法的过程中是否强制更新算法，如为True，Anylearn会对未提交的本地代码变更进行自动提交。默认为False。
    dataset_hyperparam_name : :obj:`str`, optional
        启动训练时，数据集路径作为启动命令参数传入算法的参数名。
        需指定长参数名，如 :obj:`--data` ，并省略 :obj:`--` 部分传入。
        数据集路径由Anylearn后端引擎管理。
        默认为 :obj:`dataset` 。
    dataset_id : :obj:`str`, optional
        已在Anylearn远程注册的数据集ID。
    dataset_dir : :obj:`str`, optional
        本地数据集目录路径。
        自0.17.0版本起，此参数被标记为废弃，将于0.18.0版本中移除。
        数据集上传将不再被支持。
        
        .. deprecated:: 0.17.0
            
            Uploading datasets will no longer be supported.
            remove in 0.18.0.
    dataset_archive : :obj:`str`, optional
        本地数据集压缩包路径。
        自0.17.0版本起，此参数被标记为废弃，将于0.18.0版本中移除。
        数据集上传将不再被支持。
        
        .. deprecated:: 0.17.0
            
            Uploading datasets will no longer be supported.
            remove in 0.18.0.
    dataset_cloud_names : :obj:`List[str]`, optional
        训练任务需使用的Anylearn云环境中的数据集的名称
    model_hyperparam_name : :obj:`str`, optional
        启动训练时，模型路径作为启动命令参数传入算法的参数名。
        需指定长参数名，如 :obj:`--model` ，并省略 :obj:`--` 部分传入。
        模型路径由Anylearn后端引擎管理。
        默认为 :obj:`model` 。
    model_id : :obj:`str`, optional
        已在Anylearn远程注册/转存的模型ID。
    model_cloud_names : :obj:`List[str]`, optional
        训练任务需使用的Anylearn云环境中的模型的名称
    pretrain_hyperparam_name: :obj:`str`, optional
        启动训练时，前置训练结果（间接抽象为“预训练”，即"pretrain"）路径作为启动命令参数传入算法的参数名。
        需指定长参数名，如 :obj:`--pretrain` ，并省略 :obj:`--` 部分传入。
        预训练结果路径由Anylearn后端引擎管理。
        默认为 :obj:`pretrain` 。
    pretrain_task_id: :obj:`List[str]` | :obj:`str`, optional
        在Anylearn进行过的训练的ID，一般为前缀TRAI的32位字符串。
        Anylearn会对指定的训练进行结果抽取并挂载到新一次的训练中。
    project_id : :obj:`str`, optional
        已在Anylearn远程创建的训练项目ID。
    entrypoint : :obj:`str`, optional
        启动训练的入口命令。
        自0.17.0版本起，此参数被标记为废弃，将于0.18.0版本中移除。
        请使用 :obj:`algorithm_entrypoint` 作为替代。
        
        .. deprecated:: 0.17.0
            
            use :obj:`algorithm_entrypoint` instead.
            remove in 0.18.0.
    algorithm_entrypoint : :obj:`str`, optional
        启动训练的入口命令。
    output : :obj:`str`, optional
        训练输出模型的相对路径（相对于算法目录）。
        自0.17.0版本起，此参数被标记为废弃，将于0.18.0版本中移除。
        请使用 :obj:`algorithm_output` 作为替代。
        
        .. deprecated:: 0.17.0
            
            use :obj:`algorithm_output` instead.
            remove in 0.18.0.
    algorithm_output : :obj:`str`, optional
        训练输出模型的相对路径（相对于算法目录）。
    mirror_name : :obj:`str`, optional
        训练使用的Anylearn云环境中的镜像的名称。
        自0.17.0版本起，此参数被标记为废弃，将于0.18.0版本中移除。
        请使用 :obj:`image_name` 作为替代。
        
        .. deprecated:: 0.17.0
            
            use :obj:`image_name` instead.
            remove in 0.18.0.
    image_name : :obj:`str`, optional
        训练使用的Anylearn云环境中的镜像的名称。
    resource_uploader : :obj:`ResourceUploader`, optional
        资源上传实现。
        默认使用系统内置的同步上传器 :obj:`SyncResourceUploader` 。
        自0.17.0版本起，此参数被标记为废弃，将于0.18.0版本中移除。
        上传器将不再需要。
        
        .. deprecated:: 0.17.0
            
            Uploaders will no longer be needed.
            remove in 0.18.0.
    resource_polling : :obj:`float|int`, optional
        资源上传中轮询资源状态的时间间隔（单位：秒）。
        默认为5秒。
        自0.17.0版本起，此参数被标记为废弃，将于0.18.0版本中移除。
        轮询将不再需要。
        
        .. deprecated:: 0.17.0
            
            Polling will no longer be needed.
            remove in 0.18.0.
    hyperparams : :obj:`dict`, optional
        训练超参数字典。
        超参数将作为训练启动命令的参数传入算法。
        超参数字典中的键应为长参数名，如 :obj:`--param` ，并省略 :obj:`--` 部分传入。
        如需要标识类参数（flag），可将参数的值设为空字符串，如 :obj:`{'my-flag': ''}` ，等价于 :obj:`--my-flag` 传入训练命令。
        默认为空字典。
        自0.17.0版本起，此参数被标记为废弃，将于0.18.0版本中移除。
        请使用 :obj:`algorithm_hyperparams` 作为替代。
        
        .. deprecated:: 0.17.0
            
            use :obj:`algorithm_hyperparams` instead.
            remove in 0.18.0.
    hyperparams_prefix : :obj:`str`, optional
        训练超参数键前标识，可支持hydra特殊命令行传参格式的诸如 :obj:`+key1` 、 :obj:`++key2` 、 空前置 :obj:`key3` 等需求，
        默认为 :obj:`--` 。
        自0.17.0版本起，此参数被标记为废弃，将于0.18.0版本中移除。
        请使用 :obj:`algorithm_hyperparams_prefix` 作为替代。
        
        .. deprecated:: 0.17.0
            
            use :obj:`algorithm_hyperparams_prefix` instead.
            remove in 0.18.0.
    hyperparams_delimeter :obj:`str`, optional
        训练超参数键值间的分隔符，默认为空格 :obj:` ` 。
        自0.17.0版本起，此参数被标记为废弃，将于0.18.0版本中移除。
        请使用 :obj:`algorithm_hyperparams_delimeter` 作为替代。
        
        .. deprecated:: 0.17.0
            
            use :obj:`algorithm_hyperparams_delimeter` instead.
            remove in 0.18.0.
    algorithm_hyperparams : :obj:`dict`, optional
        训练超参数字典。
        超参数将作为训练启动命令的参数传入算法。
        超参数字典中的键应为长参数名，如 :obj:`--param` ，并省略 :obj:`--` 部分传入。
        如需要标识类参数（flag），可将参数的值设为空字符串，如 :obj:`{'my-flag': ''}` ，等价于 :obj:`--my-flag` 传入训练命令。
        默认为空字典。
    algorithm_hyperparams_prefix : :obj:`str`, optional
        训练超参数键前标识，可支持hydra特殊命令行传参格式的诸如 :obj:`+key1` 、 :obj:`++key2` 、 空前置 :obj:`key3` 等需求，
        默认为 :obj:`--` 。
    algorithm_hyperparams_delimeter :obj:`str`, optional
        训练超参数键值间的分隔符，默认为空格 :obj:` ` 。
    resource_request : :obj:`List[Dict[str, Dict[str, int]]]`, optional
        训练所需计算资源的请求。
        如未填，则使用Anylearn后端的 :obj:`default` 资源组中的默认资源套餐。
        自0.17.0版本起，此参数被标记为废弃，将于0.18.0版本中移除。
        请使用 :obj:`quota_group_request` 作为替代。
        
        .. deprecated:: 0.17.0
            
            use :obj:`quota_group_request` instead.
            remove in 0.18.0.
    quota_group_name : :obj:`str`, optional
        训练所需计算资源组名称或ID。
        自0.17.0版本起，此参数被标记为废弃，将于0.18.0版本中移除。
        请在 :obj:`quota_group_request['name']` 字段中指明资源组名称作为替代。
        
        .. deprecated:: 0.17.0
            
            Please specify the quota group name in :obj:`quota_group_request['name']` instead."
            remove in 0.18.0.
    quota_group_request : :obj:`dict`, optional
        训练所需计算资源组中资源数量。
        若 :obj:`quota_group_name` 和 :obj:`quota_group_request` 有其一未填，则使用Anylearn后端的 :obj:`default` 资源组中的默认资源套餐。
    task_description : :obj:`str`, optional
        训练任务详细描述。
        若值为非空，
        且参数 :obj:`algorithm_force_update` 为 :obj:`True` 时，
        则Anylearn在自动提交本地算法变更时，
        会将此值作为commit message同步至远端
    num_nodes : :obj:`int`, optional
        分布式训练需要的节点数。
    nproc_per_node : :obj:`int`, optional
        分布式训练每个节点运行的进程数。

    Returns
    -------
    TrainTask
        创建的训练任务对象
    Algorithm
        在快速训练过程中创建或获取的算法对象
    Dataset
        在快速训练过程中创建或获取的数据集对象
    Project
        创建的训练项目对象
    """
    # Deprecations
    if algorithm_name is not None:
        cmd_warning(
            "Argument `algorithm_name` is deprecated "
            "and will be removed in v0.18.0. "
            "Please use `algorithm_cloud_name` instead."
        )
    if algorithm_dir is not None:
        cmd_warning(
            "Argument `algorithm_dir` is deprecated "
            "and will be removed in v0.18.0. "
            "Please use `algorithm_local_dir` instead."
        )
    if dataset_dir is not None:
        cmd_warning(
            "Argument `dataset_dir` is deprecated "
            "and will be removed in v0.18.0. "
            "Uploading datasets will no longer be supported in `quick_train`"
        )
    if dataset_archive is not None:
        cmd_warning(
            "Argument `dataset_archive` is deprecated "
            "and will be removed in v0.18.0. "
            "Uploading datasets will no longer be supported in `quick_train`"
        )
    if entrypoint is not None:
        cmd_warning(
            "Argument `entrypoint` is deprecated "
            "and will be removed in v0.18.0. "
            "Please use `algorithm_entrypoint` instead."
        )
    if output is not None:
        cmd_warning(
            "Argument `output` is deprecated "
            "and will be removed in v0.18.0. "
            "Please use `algorithm_output` instead."
        )
    if resource_uploader is not None:
        cmd_warning(
            "Argument `resource_uploader` is deprecated "
            "and will be removed in v0.18.0. "
            "Uploaders will no longer be needed in `quick_train`"
        )
    if resource_polling != 5:
        cmd_warning(
            "Argument `resource_polling` is deprecated "
            "and will be removed in v0.18.0. "
            "Polling will no longer be needed in `quick_train`"
        )
    if resource_request is not None:
        cmd_warning(
            "Argument `resource_request` is deprecated "
            "and will be removed in v0.18.0. "
            "Please use `quota_group_request` instead."
        )
    if quota_group_name is not None:
        cmd_warning(
            "Argument `quota_group_name` is deprecated "
            "and will be removed in v0.18.0. "
            "Please specify the quota group name in "
            "`quota_group_request['name']` instead."
        )

    # Resource request
    if quota_group_name and quota_group_request:
        resource_request = [{quota_group_name: quota_group_request}]
    elif quota_group_request and 'name' in quota_group_request:
        quota_group_name = quota_group_request.pop('name')
        resource_request = [{quota_group_name: quota_group_request}]
    resource_request = _format_resource_request(resource_request)

    # Remote model
    if model_id is None:
        model_id = []
    elif not isinstance(model_id, list):
        model_id = [model_id]
    models = [
        Model(id=_id, load_detail=True)
        for _id in model_id
    ]

    # Remote dataset
    if dataset_id is None:
        dataset_id = []
    elif not isinstance(dataset_id, list):
        dataset_id = [dataset_id]
    datasets = [
        Dataset(id=_id, load_detail=True)
        for _id in dataset_id
    ]

    # Remote pretrain task results
    if pretrain_task_id is None:
        pretrain_task_id = []
    elif not isinstance(pretrain_task_id, list):
        pretrain_task_id = [pretrain_task_id]
    pretrain_tasks = [
        TrainTask(id=_id, load_detail=True)
        for _id in pretrain_task_id 
    ]

    artifacts: List[Artifact] = []
    # Remote dataset by full names
    if dataset_cloud_names:
        artifacts.extend([
            DatasetArtifact.from_full_name(_fn)
            for _fn in dataset_cloud_names
        ])
    # Remote model by full names
    if model_cloud_names:
        artifacts.extend([
            ModelArtifact.from_full_name(_fn)
            for _fn in model_cloud_names['models']
        ])

    # Algorithm
    _algo_name = algorithm_name or algorithm_cloud_name
    _algo_dir = algorithm_dir or algorithm_local_dir
    _image_name = mirror_name if image_name == 'QUICKSTART' else image_name
    try:
        algo, current_sha = sync_algorithm(
            id=algorithm_id,
            name=_algo_name,
            dir_path=_algo_dir,
            mirror_name=_image_name,
            uploader=resource_uploader,
            polling=resource_polling,
            force=algorithm_force_update,
            commit_msg=task_description,
        )
    except AnylearnRequiredLocalCommitException:
        # Notify possible usage of algorithm_force_update=True
        raise AnylearnRequiredLocalCommitException(
            "Local algorithm code has uncommitted changes. "
            "Please commit your changes or "
            "specify `algorithm_force_update=True` "
            "to let Anylearn make an auto-commit."
    )

    # Dataset
    dset, dataset_archive, dataset_checksum = _get_or_create_dataset(
        dir_path=dataset_dir,
        archive_path=dataset_archive
    )
    if dataset_archive:
        # Local dataset registration
        _upload_dataset(dataset=dset,
                        dataset_archive=dataset_archive,
                        uploader=resource_uploader,
                        polling=resource_polling)
        DB().create_local_dataset(id=dset.id, checksum=dataset_checksum)

    if not len(datasets) and dset:
        datasets.append(dset)
    mounts = {}
    if datasets and dataset_hyperparam_name:
        mounts[dataset_hyperparam_name] = datasets
    if models and model_hyperparam_name:
        mounts[model_hyperparam_name] = models
    if pretrain_tasks and pretrain_hyperparam_name:
        mounts[pretrain_hyperparam_name] = [t.get_results_file() for t in pretrain_tasks]

    # Project
    if project_id or project_name:
        project = _get_or_create_project(project_id=project_id,
                                         project_name=project_name)
    else:
        try:
            project = _get_or_create_default_project()
        except:
            # Backward compatibility when default projects not supported
            project = _get_or_create_project()

    # Train task
    train_task_name = generate_random_name()
    _hyperparams = hyperparams or algorithm_hyperparams
    _hyperparams_prefix = hyperparams_prefix or algorithm_hyperparams_prefix
    _hyperparams_delimeter = hyperparams_delimeter or algorithm_hyperparams_delimeter
    _entrypoint = entrypoint or algorithm_entrypoint
    _output = output or algorithm_output
    train_task = _create_train_task(
        name=train_task_name,
        algorithm=algo,
        algorithm_git_ref=algorithm_git_ref or current_sha,
        project=project,
        hyperparams=_hyperparams,
        hyperparams_prefix=_hyperparams_prefix,
        hyperparams_delimeter=_hyperparams_delimeter,
        mounts=mounts,
        artifacts=artifacts,
        resource_request=resource_request,
        description=task_description,
        num_nodes=num_nodes,
        nproc_per_node=nproc_per_node,
        entrypoint=_entrypoint,
        output=_output,
        algorithm_dir=_algo_dir,
        mirror_name=_image_name,
    )

    DB().create_or_update_train_task(train_task=train_task)
    train_profile = TrainProfile(id=generate_primary_key("DESC"),
                                 train_task_id=train_task.id,
                                 entrypoint=train_task.entrypoint,
                                 algorithm_id=algo.id,
                                 dataset_id=','.join([dset.id for dset in datasets]) if datasets else None,
                                 train_params=train_task.train_params,
                                 algorithm_dir=str(algorithm_dir),
                                 dataset_dir=str(dataset_dir),
                                 dataset_archive=dataset_archive,)
    train_profile.create_in_db()
    return train_task, algo, datasets, project


def resume_unfinished_local_train_tasks():
    db = DB()
    local_list = db.get_unfinished_train_tasks()
    task_list = [TrainTask(id=local_train_task.id,
                           secret_key=local_train_task.secret_key,
                           project_id=local_train_task.project_id,
                           state=local_train_task.remote_state_sofar,
                           load_detail=True)
                 for local_train_task in local_list]
    [db.update_train_task(train_task) for train_task in task_list]
    return task_list



import logging

import click


def _fmt_msg(msg: str, level: int=logging.INFO):
    return f"[ANYLEARN] [{logging.getLevelName(level)}] {msg}"


def cmd_error(msg: str):
    click.echo(click.style(_fmt_msg(msg, logging.ERROR), fg="red"))


def cmd_warning(msg: str):
    click.echo(click.style(_fmt_msg(msg, logging.WARNING), fg="yellow"))


def cmd_success(msg: str):
    click.echo(click.style(_fmt_msg(msg, logging.INFO), fg="green"))


def cmd_info(msg: str):
    click.echo(_fmt_msg(msg, logging.INFO))



from abc import ABC, abstractmethod
import asyncio
import cgi
import time
from typing import Optional, Union

import aiofiles
import aiohttp
import requests

from anylearn.config import AnylearnConfig
from anylearn.utils.errors import AnyLearnException
from anylearn.utils.api import get_with_token, post_with_token, url_base


class ResourceDownloader(ABC):
    """
    资源下载接口
    """
    @abstractmethod
    def run(self,
            resource_id: str,
            polling: Union[float, int],
            save_path: Optional[str]=None,
           ):
        """
        执行资源下载,自定义资源上传需实现此方法

        Parameters
        ----------
        resource_id
            资源ID
        save_path
            保存路径
        """
        raise NotImplementedError


class AsyncResourceDownloader(ResourceDownloader):
    """
    资源异步下载工具类
    """

    def run(self,
            resource_id: str,
            save_path: str,
            polling: Union[float, int],
           ):
        return asyncio.run(self.__run(resource_id=resource_id,
                                      save_path=save_path,
                                      polling=polling,
                                     ))

    async def __run(self,
                    resource_id: str,
                    save_path: str,
                    polling: Union[float, int]=5,
                   ):
        """执行异步下载请求

        Args:
            resource_id: 后端资源ID
            save_path: 资源下载保存路径

        Returns:
            bool: 下载成功与否

        Raises:
            ClientResponseError: 当下载过程中出错时由aiohttp包（默认情况）抛出错误
        """

        compress_res = _compress_file(resource_id=resource_id)
        print(f"Resource compress {compress_res}")

        compress_state_res = _compress_complete(resource_id=resource_id, compression_id=compress_res['data'])
        print(f"Resource compress state {compress_state_res}")

        while not compress_state_res['data']:
            time.sleep(polling)
            compress_state_res = _compress_complete(resource_id=resource_id, compression_id=compress_res['data'])
            print(f"Resource compress state {compress_state_res}")

        headers = {'Authorization': f"Bearer {AnylearnConfig.token}"}
        async with aiohttp.ClientSession(headers=headers) as session:
            task = self.__do_download(session=session,
                                      resource_id=resource_id,
                                      compression_id=compress_res['data'],
                                      save_path=save_path)
            res = await asyncio.gather(task)
        return res

    async def __do_download(self, session: aiohttp.ClientSession,
                            resource_id: str,
                            compression_id: str,
                            save_path: str):

        res = await session.get(url=f"{url_base()}/resource/download",
                                raise_for_status=True,
                                params={
                                    'file_id': resource_id,
                                    'compression_id': compression_id,
                                    'token': AnylearnConfig.token
                                })

        content_header = res.headers.get('Content-Disposition')
        if content_header:
            _, params = cgi.parse_header(content_header)
            fileName = params['filename']
            f = await aiofiles.open(f"{save_path}/{fileName}", mode='wb')
            await f.write(await res.read())
            await f.close()
            return fileName
        else:
            return "文件下载失败"


class SyncResourceDownloader(ResourceDownloader):
    """
    资源同步下载工具类
    """

    def run(self,
            resource_id: str,
            save_path: str,
            polling: Union[float, int]=5,
           ):

        compress_res = _compress_file(resource_id=resource_id)
        print(f"Resource compress {compress_res}")

        compress_state_res = _compress_complete(resource_id=resource_id, compression_id=compress_res['data'])
        print(f"Resource compress state {compress_state_res}")

        while not compress_state_res['data']:
            time.sleep(polling)
            compress_state_res = _compress_complete(resource_id=resource_id, compression_id=compress_res['data'])
            print(f"Resource compress state {compress_state_res}")

        headers = {'Authorization': f"Bearer {AnylearnConfig.token}"}

        res = requests.get(url=f"{url_base()}/resource/download",
                           headers=headers,
                           params={
                               'file_id': resource_id,
                               'compression_id': compress_res['data'],
                               'token': AnylearnConfig.token,
                           })
        res.raise_for_status()

        content_header = res.headers.get('Content-Disposition')
        if content_header:
            _, params = cgi.parse_header(content_header)
            fileName = params['filename']
            with open(f"{save_path}/{fileName}", 'wb') as f:
                f.write(res.content)
            return fileName
        else:
            return "文件下载失败"


def _compress_file(resource_id: str):
    res = post_with_token(f"{url_base()}/resource/compression",
                          data={'file_id': resource_id})
    if not res or 'data' not in res:
        raise AnyLearnException("请求未能得到有效响应")
    return res
    
def _compress_complete(resource_id: str, compression_id: str):
    res = get_with_token(f"{url_base()}/resource/compression",
                         data={
                             'file_id': resource_id,
                             'compression_id': compression_id,
                             })
    if not res or 'data' not in res:
        raise AnyLearnException("请求未能得到有效响应")
    return res



from __future__ import annotations

from anylearn.utils.api import get_with_token, url_base
from anylearn.utils.errors import AnyLearnException


class Task:
    """
    AnyLearn训练任务类，以方法映射训练任务CRUD相关接口

    Attributes
    ----------
    id
        训练任务的唯一标识符，自动生成，由TRAI+uuid1生成的编码中后28个有效位（小写字母和数字）组成
    name
        训练任务的名称(长度1~50）)
    description
        训练任务描述（长度最大255）
    state
        训练任务状态
    creator_id
        创建者ID
    project_id
        训练项目ID
    algorithm_id
        算法ID
    algorithm_git_ref
        算法Gitea代码仓库的版本号（可以是commit号、分支名、tag名），非必填。
    hyperparams
        任务超参数列表
    hyperparams_prefix
        任务超参数键前标识
    hyperparams_delimeter
         任务超参数键值间的分隔符
    input_artifact_ids
        任务输入资产ID列表（以逗号分隔）
    output_artifact_id
        任务输出文件资产ID
    secret_key
        密钥
    created_at
        创建时间
    started_at
        任务开始执行时间
    finished_at
        结束时间
    envs
        训练任务环境变量（默认''）
    hpo
        是否开启超参数自动调优
    hpo_search_space
        开启超参数自动调优时不能为空
    final_metric
        最终结果指标
    resource_request : :obj:`List[Dict[str, Dict[str, int]]]`, optional
        训练所需计算资源的请求。
        如未填，则使用Anylearn后端的:obj:`default`资源组中的默认资源套餐。
    entrypoint
        算法训练的启动命令，非标准算法必填
    output_path
        算法训练结果（模型）存储目录路径，非标准算法必填
    image_id
        训练使用的镜像ID，默认为空，即使用算法绑定的镜像ID
    """

    def __init__(self, *_, **kwargs):
        self.id = kwargs.pop('id', None)
        self.name = kwargs.pop('name', None)
        self.description = kwargs.pop('description', None)
        self.state = kwargs.pop('state', None)
        self.creator_id = kwargs.pop('creator_id', None)
        self.project_id = kwargs.pop('project_id', None)
        self.algorithm_id = kwargs.pop('algorithm_id', None)
        self.algorithm_git_ref = kwargs.pop('algorithm_git_ref', None)
        self.hyperparams = kwargs.pop('train_params', None)
        self.hyperparams_prefix = kwargs.pop('train_params_prefix', "--")
        self.hyperparams_delimeter = kwargs.pop('train_params_delimeter', " ")
        self.input_artifact_ids = kwargs.pop('files', None)
        self.output_artifact_id = kwargs.pop('results_id', None)
        self.secret_key = kwargs.pop('secret_key', None)
        self.created_at = kwargs.pop('create_time', None)
        self.started_at = kwargs.pop('start_time', None)
        self.finished_at = kwargs.pop('finish_time', None)
        self.envs = kwargs.pop('envs', None)
        self.hpo = kwargs.pop('hpo', False)
        self.hpo_search_space = kwargs.pop('hpo_search_space', None)
        self.final_metric = kwargs.pop('final_metric', None)
        self.resource_request = kwargs.pop('resource_request', None)
        self.distr_num_nodes = kwargs.pop('num_nodes', 1)
        self.distr_num_proc = kwargs.pop('nproc_per_node', 1)
        self.entrypoint = kwargs.pop('entrypoint', None)
        self.output_path = kwargs.pop('output', None)
        self.image_id = kwargs.pop('mirror_id', None)

    @classmethod
    def from_id(cls, id_: str) -> Task:
        res = get_with_token(
            f"{url_base()}/train_task/query",
            params={'id': id_},
        )
        if not res or not isinstance(res, list):
            raise AnyLearnException("Request failed")
        return Task(**res[0])



import os
from pathlib import Path
from typing import Union

from anylearn.utils.errors import AnyLearnException


def ensure_dir(dir_: Union[os.PathLike, str]):
    root_dir = Path(dir_)
    if root_dir.exists() and root_dir.is_file():
        raise AnyLearnException(
            f"Path `{str(dir_)}` is expected to be a directory, "
            "a file given"
        )
    root_dir.mkdir(parents=True, exist_ok=True)



from __future__ import annotations
from datetime import datetime
from typing import Optional, Union

from anylearn.utils.api import url_base, get_with_token
from anylearn.utils.errors import AnyLearnException
from anylearn.interfaces.resource.resource import Resource

class Dataset(Resource):
    """
    AnyLearn数据集类，以方法映射数据集CRUD相关接口
        
    Attributes
    ----------
    id
        数据集的唯一标识符，自动生成，由ALGO+uuid1生成的编码中后28个有效位（小写字母和数字）组成（自动生成）
    name
        数据集的名称（长度1~50）
    description
        数据集描述（长度最大255）
    state
        数据集状态
    public
        数据集是否公开（默认为False）
    upload_time
        数据集上传时间
    filename
        下一步中会被分片上传的文件的完整文件名（包括扩展名）（长度1~128）
    is_zipfile
        是否为zip文件
    file_path
        数据集文件路径
    size
        数据集文件大小
    creator_id
        创建者ID
    node_id
        数据集节点ID
    load_detail
        初始化时是否加载详情
    """

    """具体资源信息配置"""
    _fields = {
        # 资源创建/更新请求包体中必须包含且不能为空的字段
        'required': {
            'create': ['name'],
            'update': ['id', 'name'],
        },
        # 资源创建/更新请求包体中包含的所有字段
        'payload': {
            'create': ['name', 'description', 'public',
                       'filename'],
            'update': ['id', 'name', 'description', 'public',
                       'filename'],
        },
    }
    """
    创建/更新对象时：

    - 必须包含且不能为空的字段 :obj:`_fields['required']`
    - 所有字段 :obj:`_fields['payload']`
    """

    def __init__(self,
                 id: Optional[str]=None,
                 name: Optional[str]=None,
                 description: Optional[str]=None,
                 state: Optional[int]=None,
                 public=False,
                 upload_time: Optional[Union[datetime, str]]=None,
                 filename: Optional[str]=None,
                 is_zipfile: Optional[bool]=None,
                 file_path: Optional[str]=None,
                 size: Optional[str]=None,
                 creator_id: Optional[str]=None,
                 node_id: Optional[str]=None,
                 load_detail=False):
        """
        Parameters
        ----------
        id
            数据集的唯一标识符，自动生成，由ALGO+uuid1生成的编码中后28个有效位（小写字母和数字）组成（自动生成）
        name
            数据集的名称（长度1~50）
        description
            数据集描述（长度最大255）
        state
            数据集状态
        public
            数据集是否公开（默认为False）
        upload_time
            数据集上传时间
        filename
            下一步中会被分片上传的文件的完整文件名（包括扩展名）（长度1~128）
        is_zipfile
            是否为zip文件
        file_path
            数据集文件路径
        size
            数据集文件大小
        creator_id
            创建者ID
        node_id
            数据集节点ID
        load_detail
            初始化时是否加载详情
        """
        super().__init__(id=id, name=name, description=description,
                         state=state, public=public,
                         upload_time=upload_time, filename=filename,
                         is_zipfile=is_zipfile, file_path=file_path, size=size,
                         creator_id=creator_id, node_id=node_id,
                         load_detail=load_detail)

    def __eq__(self, other: Dataset) -> bool:
        if not isinstance(other, Dataset):
            return NotImplemented
        return all([
            self.id == other.id,
            self.name == other.name,
            self.description == other.description,
            self.public == other.public,
            # self.state == other.state,
            # self.upload_time == other.upload_time,
            # self.filename == other.filename,
            # self.file_path == other.file_path,
            # self.is_zipfile == other.is_zipfile,
            # self.size == other.size,
            # self.creator_id == other.creator_id,
            # self.node_id == other.node_id,
        ])

    @classmethod
    def get_list(cls) -> list:
        """
        获取数据集列表
        
        Returns
        -------
        List [Dataset]
            数据集对象的集合。
        """
        res = get_with_token(f"{url_base()}/dataset/list")
        if res is None or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        return [
            Dataset(id=d['id'], name=d['name'], description=d['description'],
                    state=d['state'], public=d['public'],
                    upload_time=d['upload_time'])
            for d in res
        ]

    def get_detail(self):
        """
        获取数据集详细信息

        - 对象属性 :obj:`id` 应为非空

        Returns
        -------
        Dataset
            数据集对象。
        """
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/dataset/query",
                             params={'id': self.id})
        if not res or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        res = res[0]
        self.__init__(id=res['id'], name=res['name'],
                      description=res['description'], state=res['state'],
                      public=res['public'],
                      upload_time=res['upload_time'], filename=res['filename'],
                      is_zipfile=True if res['is_zipfile'] == 1 else False,
                      file_path=res['file'], size=res['size'],
                      creator_id=res['creator_id'], node_id=res['node_id'])

    def _namespace(self):
        return "dataset"



class InvalidArguments(Exception):
    def __init__(self, message='some arguments are invalid'):
        super(InvalidArguments, self).__init__()
        self.message = message
    
    def __str__(self, *args, **kwargs):
        return self.message


class AnyLearnException(Exception):
    """AnyLearnSDK通用异常类"""

    def __init__(self, msg):
        """
        :param msg: 错误信息
        """
        self.msg = msg

    def __str__(self):
        s = f"错误信息: {self.msg}"
        return s

    def __repr__(self):
        __repr = f"[{self.__class__.__name__}] {self.msg}"
        return __repr


class AnyLearnAuthException(AnyLearnException):
    """AnyLearnSDK鉴权异常类"""

    def __init__(self, msg="无效的用户身份信息或令牌"):
        super().__init__(msg)


class AnyLearnMissingParamException(AnyLearnException):
    """AnyLearnSDK接口调用参数缺失异常类"""

    def __init__(self, msg="缺少参数或参数格式不正确"):
        super().__init__(msg)


class AnyLearnNotSupportedException(AnyLearnException):
    """AnyLearnSDK接口不支持异常类"""

    def __init__(self, msg="暂不支持该操作"):
        super().__init__(msg)


class AnylearnRequiredLocalCommitException(AnyLearnException):
    """本地算法目录有未主动提交的变更"""

    DEFAULT_MESSAGE = (
        "Algorithm dir is not clean, "
        "please commit your changes first"
    )

    def __init__(self, msg=DEFAULT_MESSAGE):
        super().__init__(msg)



from datetime import datetime as dt
from multiprocessing import Queue
from pathlib import Path
import sys
from threading import Thread
import time
import traceback
from typing import Optional, Tuple, Union
from urllib.parse import quote, urlparse

from .utils import (
    _get_or_create_resource_archive,
    get_mirror_by_name,
    make_name_by_path,
)
from ..config import AnylearnConfig
from ..interfaces.resource import (
    Algorithm,
    Resource,
    ResourceState,
    ResourceUploader,
    SyncResourceUploader,
)
from ..utils import logger
from ..utils.errors import (
    AnyLearnException,
    AnyLearnMissingParamException,
    AnyLearnNotSupportedException,
    AnylearnRequiredLocalCommitException,
)


def sync_algorithm(
    id: Optional[str]=None,
    name: Optional[str]=None,
    dir_path: Optional[Union[str, Path]]=None,
    mirror_name: Optional[str]="QUICKSTART",
    uploader: Optional[ResourceUploader]=None,
    polling: Union[float, int]=5,
    force: bool=False,
    commit_msg: str=None,
) -> Tuple[Algorithm, str]:
    if id:
        return _sync_remote_algorithm(
            id=id,
            image_name=mirror_name,
        )
    else:
        return _sync_local_algorithm(
            name=name,
            dir_path=dir_path,
            image_name=mirror_name,
            uploader=uploader,
            polling=polling,
            force=force,
            commit_msg=commit_msg,
        )


def _sync_remote_algorithm(
    id: str,
    image_name: Optional[str]='QUICKSTART',
) -> Tuple[Algorithm, Optional[str]]:
    algo = Algorithm(id=id, load_detail=True)
    algo.mirror_id = get_mirror_by_name(image_name).id
    try:
        algo.save()
    except:
        logger.warning(
            f"Failed to update algorithm {id}. "
            "Training will still be submitted and "
            "the algorithm will be used as is. "
            "Please make sure you have write-permission to this algorithm. "
        )
    algo.get_detail()
    return algo, None


def _sync_local_algorithm(
    name: str,
    dir_path: Union[str, Path],
    image_name: str="QUICKSTART",
    uploader: Optional[ResourceUploader]=None,
    polling: Union[float, int]=5,
    force: bool=False,
    commit_msg: str=None,
) -> Tuple[Algorithm, Optional[str]]:
    # Required params
    if not dir_path:
        raise AnyLearnMissingParamException((
            "Parameters "
            "`dir_path`"
            "is required"
        ))
    if not name:
        name = make_name_by_path(dir_path)
    _check_local_algorithm(dir_path)
    algo = _get_or_create_algorithm(
        name=name,
        dir_path=dir_path,
        image_name=image_name,
    )
    if AnylearnConfig.git_ready and algo.git_address:
        current_sha = _sync_algorithm_repo(
            algorithm=algo,
            dir_path=dir_path,
            force=force,
            commit_msg=commit_msg,
        )
        logger.info("Waiting algorithm to be ready...")
        algo.sync_finish(checkout_sha=current_sha)
    else:
        _upload_algorithm(
            algorithm=algo,
            dir_path=dir_path,
            uploader=uploader,
            polling=polling,
        )
        current_sha = None
    return algo, current_sha


def _check_local_algorithm(dir_path: str) -> None:
    logger.info(f"Verifying local algorithm in '{dir_path}'")
    dir_path = Path(dir_path)
    # Check directory
    if not all([
        dir_path.exists(),
        dir_path.is_dir(),
    ]):
        raise AnyLearnException(
            "Parameter `dir_path` must be an existing directory"
        )
    # Check requirements.txt
    if not (dir_path / "requirements.txt").exists():
        raise AnyLearnException(("Missing 'requirements.txt' "
                                 "in algorithm directory"))


def _get_or_create_algorithm(name: str,
                             dir_path: Union[str, Path],
                             image_name: str="QUICKSTART") -> Algorithm:
    # Algo name or dir name
    if not name:
        logger.warning(
            "Algorithm `name` is not set, using directory basename"
        )
        name = dir_path.name
    image = get_mirror_by_name(image_name)
    algo = _get_algorithm_by_name(name)
    if algo:
        logger.warning(f"Algorithm named '{name}' already exists.")
    else:
        logger.info(f"Creating new algorithm '{name}'")
        algo =_create_new_algorithm(name)
    # Set execution metadata
    algo.mirror_id = image.id
    # Existing/new algo needs respectively create/update towards remote
    algo.save()
    algo.get_detail()
    return algo


def _get_algorithm_by_name(name) -> Optional[Algorithm]:
    try:
        return Algorithm.get_user_custom_algorithm_by_name(name=name)
    except:
        return None


def _create_new_algorithm(name: str) -> Algorithm:
    algo = Algorithm(
        name=name,
        description="SDK_QUICKSTART",
        public=False,
        filename=f"{name}.zip", # legacy
        is_zipfile=True, # legacy
        follows_anylearn_norm=False,
    )
    return algo


def _sync_algorithm_repo(
    algorithm: Algorithm,
    dir_path: str,
    force: bool=False,
    commit_msg: str=None,
) -> None:
    if AnylearnConfig.git_ready:
        from git import Repo
        from .git_progress_printer import GitProgressPrinter
    logger.info(f"Synchronizing algorithm repo to {algorithm.git_address}")
    repo = Repo.init(dir_path)
    # Origin
    try:
        origin = repo.remotes.anylearn
        origin.set_url(_auth_git_url(algorithm.git_address))
    except AttributeError:
        origin = repo.create_remote(
            'anylearn',
            _auth_git_url(algorithm.git_address),
        )
    # Try fetching
    if origin.exists():
        logger.debug("Fetching")
        origin.fetch()
    else:
        logger.debug("Skipped `git fetch`")
    # add & commit & push
    logger.debug("Staging")
    repo.git.add(A=True)
    try:
        # Check staged changes
        empty_stage = not repo.index.diff(repo.tree())
    except ValueError:
        # Non published repo occurs ValueError
        empty_stage = False
    if not empty_stage:
        if not force:
            raise AnylearnRequiredLocalCommitException
        logger.warning(
            f"Forced auto-committing changes in '{dir_path}'"
        )
        logger.debug("Commiting")
        if not commit_msg:
            commit_msg = _commit_msg()
        repo.index.commit(commit_msg)
    push_args = {'all': True}
    if force:
        push_args['force'] = True
    logger.debug("Pushing")
    origin.push(progress=GitProgressPrinter(), **push_args)
    # Reset remote URL without credentials
    origin.set_url(algorithm.git_address)
    return repo.head.commit.hexsha


def _auth_git_url(url):
    p = urlparse(url)
    root = p.netloc.split('@')[-1]
    auth = f"{AnylearnConfig.username}:{quote(AnylearnConfig.password)}"
    return p._replace(netloc=f"{auth}@{root}").geturl()


def _commit_msg():
    return (
        "Anylearn auto-commit "
        f"{dt.now().strftime('%Y-%m-%d %H:%M:%S')}"
    )


def _upload_algorithm(algorithm: Algorithm,
                      dir_path: str,
                      uploader: Optional[ResourceUploader]=None,
                      polling: Union[float, int]=5):
    if not uploader:
        uploader = SyncResourceUploader()
    archive = _get_or_create_resource_archive(
        name=algorithm.name,
        dir_path=dir_path,
    )
    q = Queue()
    t_algorithm = Thread(
        target=__do_upload,
        args=[q],
        kwargs={
            'resource_id': algorithm.id,
            'file_path': archive,
            'uploader': uploader,
        }
    )
    logger.info(f"Uploading algorithm {algorithm.name}...")
    t_algorithm.start()
    err = q.get()
    t_algorithm.join()
    Path(archive).unlink()
    if err:
        ex_type, ex_value, tb_str = err
        message = f"{str(ex_value)} (in subprocess)\n{tb_str}"
        raise ex_type(message)
    __wait_algorithm_ready(algorithm, polling)
    logger.info("Successfully uploaded algorithm")


def __do_upload(q: Queue, *args, **kwargs):
    try:
        Resource.upload_file(*args, **kwargs)
        err = None
    except:
        ex_type, ex_value, tb = sys.exc_info()
        err = ex_type, ex_value, ''.join(traceback.format_tb(tb))
    q.put(err)


def __wait_algorithm_ready(algorithm: Algorithm, polling: Union[float, int]=5):
    algorithm.state = None
    finished = [ResourceState.ERROR, ResourceState.READY]
    while algorithm.state not in finished:
        time.sleep(polling)
        algorithm.get_detail()
    if algorithm.state == ResourceState.ERROR:
        raise AnyLearnException("An error occured when uploading algorithm")
    logger.info("Algorithm is ready")



from __future__ import annotations
from datetime import datetime
import json
from typing import Optional, Union
import re

from anylearn.utils.api import url_base, get_with_token, post_with_token
from anylearn.utils.errors import AnyLearnException
from anylearn.interfaces.resource.resource import Resource, ResourceState

class Algorithm(Resource):
    """
    AnyLearn算法类，以方法映射算法CRUD相关接口

    Attributes
    ----------
    id
        算法的唯一标识符，自动生成，由ALGO+uuid1生成的编码中后28个有效位（小写字母和数字）组成）（自动生成）
    name
        算法的名称
    description
        算法的描述
    state
        算法状态
    public
        算法是否公开（默认为False）
    upload_time
        算法上传时间
    filename
        下一步中会被分片上传的文件的完整文件名（包括扩展名）（非空 长度1~128）
    is_zipfile
        是否为zip文件
    file_path
        算法文件路径
    size
        算法文件大小
    creator_id
        算法的创建者
    node_id
        算法节点ID
    tags
        算法的标签
    mirror_id
        算法使用的基础镜像的id
    train_params
        算法的训练参数，包括数据集参数
    follows_anylearn_norm
        算法是否符合Anylearn的算法规范（默认为True）
    git_address
        算法的Anylearn Gitea远端代码仓库地址
    git_migrated
        算法是否由传统模式自动迁移至Anylearn Gitea
    load_detail
        初始化时是否加载详情
    """

    """具体资源信息配置"""
    _fields = {
        # 资源创建/更新请求包体中必须包含且不能为空的字段
        'required': {
            'create': ['name', 'mirror_id'],
            'update': ['id', 'name'],
        },
        # 资源创建/更新请求包体中包含的所有字段
        'payload': {
            'create': ['name', 'description', 'public',
                       'filename', 'tags', 'mirror_id', 'train_params',
                       'follows_anylearn_norm'],
            'update': ['id', 'name', 'description', 'public', 
                       'tags', 'mirror_id', 'train_params',
                       'follows_anylearn_norm', 'filename'],
        },
    }
    """
    创建/更新对象时：

    - 必须包含且不能为空的字段 :obj:`_fields['required']`
    - 所有字段 :obj:`_fields['payload']`
    """

    __train_params = None
    required_train_params = []
    default_train_params = {}

    def __init__(self,
                 id: Optional[str]=None,
                 name: Optional[str]=None,
                 description: Optional[str]=None,
                 state: Optional[int]=None,
                 public=False,
                 upload_time: Optional[Union[datetime, str]]=None,
                 filename: Optional[str]=None,
                 is_zipfile: Optional[int]=None,
                 file_path: Optional[str]=None,
                 size: Optional[str]=None,
                 creator_id: Optional[str]=None,
                 node_id: Optional[str]=None,
                 tags: Optional[str]=None,
                 mirror_id: Optional[str]=None,
                 train_params: Optional[str]=None,
                 follows_anylearn_norm=True,
                 git_address: Optional[str]=None,
                 git_migrated: Optional[bool]=False,
                 load_detail=False):
        """
        Parameters
        ----------
        id
            算法的唯一标识符，自动生成，由ALGO+uuid1生成的编码中后28个有效位（小写字母和数字）组成）（自动生成）
        name
            算法的名称
        description
            算法的描述
        state
            算法状态
        public
            算法是否公开（默认为False）
        upload_time
            算法上传时间
        filename
            下一步中会被分片上传的文件的完整文件名（包括扩展名）（非空 长度1~128）
        is_zipfile
            是否为zip文件
        file_path
            算法文件路径
        size
            算法文件大小
        creator_id
            算法的创建者
        node_id
            算法节点ID
        tags
            算法的标签
        mirror_id
            算法使用的基础镜像的id
        train_params
            算法的训练参数，包括数据集参数
        follows_anylearn_norm
            算法是否符合Anylearn的算法规范（默认为True）
        git_address
            算法的Anylearn Gitea远端代码仓库地址
        git_migrated
            算法是否由传统模式自动迁移至Anylearn Gitea
        load_detail
            初始化时是否加载详情
        """
        self.tags = tags
        self.mirror_id = mirror_id
        self.train_params = train_params
        self.follows_anylearn_norm = follows_anylearn_norm
        self.git_address = git_address
        self.git_migrated = git_migrated
        super().__init__(id=id, name=name, description=description,
                         state=state, public=public,
                         upload_time=upload_time, filename=filename,
                         is_zipfile=is_zipfile, file_path=file_path, size=size,
                         creator_id=creator_id, node_id=node_id,
                         load_detail=load_detail)

    def __eq__(self, other: Algorithm) -> bool:
        if not isinstance(other, Algorithm):
            return NotImplemented
        return all([
            self.id == other.id,
            self.name == other.name,
            self.description == other.description,
            self.public == other.public,
            # self.state == other.state,
            # self.upload_time == other.upload_time,
            # self.filename == other.filename,
            # self.file_path == other.file_path,
            # self.is_zipfile == other.is_zipfile,
            # self.size == other.size,
            # self.creator_id == other.creator_id,
            # self.node_id == other.node_id,
            # self.tags == other.tags,
            # self.mirror_id == other.mirror_id,
            # self.train_params == other.train_params,
            self.follows_anylearn_norm == other.follows_anylearn_norm,
            self.git_address == other.git_address,
        ])

    @classmethod
    def get_list(cls) -> list:
        """
        获取算法列表
        
        Returns
        -------
        List [Algorithm]
            算法对象的集合。
        """
        res = get_with_token(f"{url_base()}/algorithm/list")
        if res is None or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        return [
            Algorithm(id=a['id'], name=a['name'], description=a['description'],
                      state=a['state'], public=a['public'],
                      upload_time=a['upload_time'], tags=a['tags'],
                      follows_anylearn_norm=a['follows_anylearn_norm'],
                      git_address=a['git_address'],
                      git_migrated=a['git_migrated'])
            for a in res
        ]

    def get_detail(self):
        """
        获取算法详细信息

        - 对象属性 :obj:`id` 应为非空

        Returns
        -------
        Algorithm
            算法对象。
        """
        self._check_fields(required=['id'])
        res = get_with_token(f"{url_base()}/algorithm/query",
                             params={'id': self.id})
        if not res or not isinstance(res, list):
            raise AnyLearnException("请求未能得到有效响应")
        res = res[0]
        self.__init__(id=res['id'], name=res['name'],
                      description=res['description'], state=res['state'],
                      public=res['public'],
                      upload_time=res['upload_time'], filename=res['filename'],
                      is_zipfile=True if res['is_zipfile'] == 1 else False,
                      file_path=res['file'], size=res['size'],
                      creator_id=res['creator_id'], node_id=res['node_id'],
                      mirror_id=res['mirror_id'], tags=res['tags'],
                      train_params=res['train_params'],
                      follows_anylearn_norm=res['follows_anylearn_norm'],
                      git_address=res['git_address'],
                      git_migrated=res['git_migrated'])

    def _namespace(self):
        return "algorithm"

    # def _create(self):
    #     if not self.follows_anylearn_norm:
    #         self._check_fields(required=['entrypoint_training',
    #                                      'output_training'])
    #     return super()._create()

    def _payload_create(self):
        payload = super()._payload_create()
        # Algorithm name verification
        self.__validate_algo_name(payload)
        payload['follows_anylearn_norm'] = int(payload['follows_anylearn_norm'])
        return self.__payload_algo_params_to_str(payload)

    def _payload_update(self):
        payload = super()._payload_update()
        # Algorithm name verification
        self.__validate_algo_name(payload)
        payload['follows_anylearn_norm'] = int(payload['follows_anylearn_norm'])
        return self.__payload_algo_params_to_str(payload)

    def __validate_algo_name(self, payload):
        name_pattern = re.compile(r'[a-zA-Z0-9_.-]+$')
        if not name_pattern.match(payload['name']):
            raise AnyLearnException(
                "Algorithm name should contain only alphanumeric, dash ('-'), "
                "underscore ('_') and dot ('.') characters"
            )

    def __payload_algo_params_to_str(self, payload):
        if 'train_params' in payload:
            payload['train_params'] = json.dumps(payload['train_params'])
        if 'owner' in payload and isinstance(payload['owner'], list):
            payload['owner'] = ",".join(payload['owner'])
        return payload

    @property
    def train_params(self):
        """
        获取训练参数
        """
        return self.__train_params

    @train_params.setter
    def train_params(self, train_params):
        """
        设置训练参数
        """
        if not train_params:
            return
        params = json.loads(train_params)
        if params == None:
            return
        self.__train_params = params
        (self.required_train_params,
         self.default_train_params) = self.__parse_params(params)

    def __parse_params(self, params):
        required_params = [p for p in params if 'default' not in p]
        default_params = {p['name']: p['default'] for p in params
                          if 'default' in p}
        return required_params, default_params

    @classmethod
    def get_user_custom_algorithm_by_name(cls, name: str):
        """
        根据算法名称获取当前用户的自定义算法

        Parameters
        ----------
        name : :obj:`str`
            算法名称。

        Returns
        -------
        Algorithm
            算法对象。
        """
        res = get_with_token(f"{url_base()}/algorithm/custom",
                             params={'name': name})
        if not res or not isinstance(res, dict):
            raise AnyLearnException("请求未能得到有效响应")
        return Algorithm(
            id=res['id'],
            name=res['name'],
            description=res['description'],
            state=res['state'],
            public=res['public'],
            upload_time=res['upload_time'],
            filename=res['filename'],
            is_zipfile=True if res['is_zipfile'] == 1 else False,
            file_path=res['file'],
            size=res['size'],
            creator_id=res['creator_id'],
            node_id=res['node_id'],
            tags=res['tags'],
            mirror_id=res['mirror_id'],
            train_params=res['train_params'],
            follows_anylearn_norm=res['follows_anylearn_norm'],
            git_address=res['git_address'],
            git_migrated=res['git_migrated'],
        )

    def sync_finish(self, checkout_sha:str=None):
        """
        算法代码仓库同步完成接口

        - 对象属性 :obj:`id` 应为非空

        Returns
        -------
        None
        """
        self._check_fields(required=['id'])
        post_with_token(f"{url_base()}/resource/sync_finish",
                        data={
                            'algorithm_id': self.id,
                            'checkout_sha': checkout_sha,
                        })
        self.get_detail()
        if self.state != ResourceState.READY:
            raise AnyLearnException("Algorithm is not ready after sync finish")



from __future__ import annotations
from datetime import datetime
import json
from threading import Thread
from typing import Dict, List, Optional, Union
from urllib.parse import urlparse, ParseResult

from anylearn.config import AnylearnConfig
from anylearn.interfaces import BaseObject, Project, TrainTask
from anylearn.interfaces.resource import (
    Algorithm,
    AsyncResourceDownloader,
    Dataset,
    Model,
    Resource,
    ResourceDownloader,
)
from anylearn.utils import logger, utc_plus_8
from anylearn.utils.api import url_base, get_with_token, post_with_token
from anylearn.utils.errors import AnyLearnException


class HpoExperiment(BaseObject):
    """
    调参实验描述类。

    一个调参实验的描述包括了
    调参任务配置、
    训练配置以及算法、
    数据集、
    训练的元信息
    等几方面，
    用户可以根据这些信息可以在多次调参实验之间进行区别和回忆。
    调参实验描述中封装了一系列与实验相关的功能，
    如：获取全部实验、获取某一实验、停止实验、重启实验、输出最佳模型等等，
    以便对任一调参实验的生命周期进行完整的回溯和管理。

    Attributes
    ----------
    project_id : :obj:`str`
        已在Anylearn远程创建的训练项目ID。
    algorithm_id : :obj:`str`
        已在Anylearn远程注册的算法ID。
    dataset_id : :obj:`str`
        已在Anylearn远程注册的数据集ID。
    hpo_search_space : :obj:`dict`
        超参数搜索空间，格式详询 :obj:`NNI` 文档。
    hpo_max_runs : :obj:`int`, optional
        最大调优任务数，达成即终止实验。默认为 :obj:`10` 。
    hpo_max_duration : :obj:`str`, optional
        最长调优总时间，达成即终止实验。
        格式为 :obj:`数字 + s|m|h|d`，
        例如： 半天 :obj:`0.5d` ， 十分钟 :obj:`10m` 。
        默认为 :obj:`24h` 。
    hpo_tuner_name : :obj:`str`, optional
        调优算法名称。
        详见 `NNI内置调优算法 <https://nni.readthedocs.io/zh/stable/builtin_tuner.html>`_ 。
        默认为 :obj:`TPE` 。
    hpo_mode : :obj:`str`, optional
        调优模式，即，最大化或最小化指标。
        可选项： :obj:`maximize` 或 :obj:`minimize` 。
    hpo_concurrency : :obj:`int`, optional
        调优期望并行度，即期望同时运行的任务数，实际并行度视后端资源调度情况而定。
        默认为 :obj:`1` 。
    resource_request : :obj:`List[Dict[str, Dict[str, int]]]`, optional
        单次训练所需计算资源的请求。
        如未填，则使用Anylearn后端的:obj:`default`资源组中的默认资源套餐。
    created_at : :obj:`str`, optional
        调参实验的创建时间。
        默认为东八时区的当前时间。
    hpo_id : :obj:`str`, optional
        调参实验ID。
    hpo_ip : :obj:`str`, optional
        调参实验容器IP地址。
    hpo_port : :obj:`int`, optional
        调优实验的本地端口绑定。
    project: :obj:`Project`, optional
        已在Anylearn远程创建的训练项目对象
    algorithm : :obj:`Algorithm`, optional
        已在Anylearn远程注册的算法对象。
    dataset : :obj:`Dataset`, optional
        已在Anylearn远程注册的数据集对象。
    dataset_hyperparam_name : :obj:`str`, optional
        启动训练时，数据集路径作为启动命令参数传入算法的参数名。
        需指定长参数名，如 :obj:`--data` ，并省略 :obj:`--` 部分传入。
        数据集路径由Anylearn后端引擎管理。
        默认为 :obj:`dataset` 。
    tasks: :obj:`dict`, optional
        调参实验中的训练任务ID与调参任务ID的映射。
    err: :obj:`list`, optional
        调参准备过程中出现的错误日志。
    """

    _fields = {
        # 资源创建/更新请求包体中必须包含且不能为空的字段
        'required': {
            'create': ['project_id', 'algorithm_id', 'dataset_id',
                       'hpo_search_space', 'dataset_hyperparam_name'],
            'update': [],
        },
        # 资源创建/更新请求包体中包含的所有字段
        'payload': {
            'create': ['project_id', 'algorithm_id', 'dataset_id',
                       'hpo_search_space', 'hpo_max_runs', 'hpo_max_duration',
                       'hpo_tuner_name', 'hpo_mode', 'hpo_concurrency',
                       'dataset_hyperparam_name', 'resource_request'],
            'update': [],
        },
    }

    def __init__(self,
                 project_id: str,
                 algorithm_id: Optional[str]=None,
                 dataset_id: Optional[str]=None,
                 hpo_search_space: Optional[dict]=None,
                 hpo_max_runs: Optional[int]=10,
                 hpo_max_duration: Optional[int]="24h",
                 hpo_tuner_name: Optional[str]="TPE",
                 hpo_mode: Optional[str]="maximize",
                 hpo_concurrency: Optional[int]=1,
                 resource_request: Optional[List[Dict[str, Dict[str, int]]]]=None,
                 created_at: Optional[datetime]=utc_plus_8(),
                 hpo_id: Optional[str]=None,
                 hpo_ip: Optional[str]=None,
                 hpo_port: Optional[int]=None,
                 hpo_status: Optional[str]=None,
                 project: Optional[Project]=None,
                 algorithm: Optional[Algorithm]=None,
                 dataset: Optional[Dataset]=None,
                 dataset_hyperparam_name: Optional[str]='dataset',
                 tasks: Optional[dict]=dict(),
                 err: Optional[list]=None,
                 load_detail: bool=False):
        """
        Parameters
        ----------
        project_id : :obj:`str`
            已在Anylearn远程创建的训练项目ID。
        algorithm_id : :obj:`str`
            已在Anylearn远程注册的算法ID。
        dataset_id : :obj:`str`
            已在Anylearn远程注册的数据集ID。
        hpo_search_space : :obj:`dict`
            超参数搜索空间，格式详询 :obj:`NNI` 文档。
        hpo_max_runs : :obj:`int`, optional
            最大调优任务数，达成即终止实验。默认为 :obj:`10` 。
        hpo_max_duration : :obj:`str`, optional
            最长调优总时间，达成即终止实验。
            格式为 :obj:`数字 + s|m|h|d`，
            例如： 半天 :obj:`0.5d` ， 十分钟 :obj:`10m` 。
            默认为 :obj:`24h` 。
        hpo_tuner_name : :obj:`str`, optional
            调优算法名称。
            详见 `NNI内置调优算法 <https://nni.readthedocs.io/zh/stable/builtin_tuner.html>`_ 。
            默认为 :obj:`TPE` 。
        hpo_mode : :obj:`str`, optional
            调优模式，即，最大化或最小化指标。
            可选项： :obj:`maximize` 或 :obj:`minimize` 。
        hpo_concurrency : :obj:`int`, optional
            调优期望并行度，即期望同时运行的任务数，实际并行度视后端资源调度情况而定。
            默认为 :obj:`1` 。
        resource_request : :obj:`List[Dict[str, Dict[str, int]]]`, optional
            单次训练所需计算资源的请求。
            如未填，则使用Anylearn后端的:obj:`default`资源组中的默认资源套餐。
        created_at : :obj:`str`, optional
            调参实验的创建时间。
            默认为东八时区的当前时间。
        hpo_id : :obj:`str`, optional
            调参实验ID。
        hpo_ip : :obj:`str`, optional
            调参实验容器IP地址。
        hpo_port : :obj:`int`, optional
            调优实验的本地端口绑定。
        project: :obj:`Project`, optional
            已在Anylearn远程创建的训练项目对象
        algorithm : :obj:`Algorithm`, optional
            已在Anylearn远程注册的算法对象。
        dataset : :obj:`Dataset`, optional
            已在Anylearn远程注册的数据集对象。
        dataset_hyperparam_name : :obj:`str`, optional
            启动训练时，数据集路径作为启动命令参数传入算法的参数名。
            需指定长参数名，如 :obj:`--data` ，并省略 :obj:`--` 部分传入。
            数据集路径由Anylearn后端引擎管理。
            默认为 :obj:`dataset` 。
        tasks: :obj:`dict`, optional
            调参实验中的训练任务ID与调参任务ID的映射。
        err: :obj:`list`, optional
            调参准备过程中出现的错误日志。
        """
        self.project_id = project_id
        self.algorithm_id = algorithm_id
        self.dataset_id = dataset_id
        self.hpo_search_space = hpo_search_space
        self.hpo_max_runs = hpo_max_runs
        self.hpo_max_duration = hpo_max_duration
        self.hpo_tuner_name = hpo_tuner_name
        self.hpo_mode = hpo_mode
        self.hpo_concurrency = hpo_concurrency
        self.resource_request = resource_request
        self.created_at = created_at
        self.hpo_id = hpo_id
        self.hpo_ip = hpo_ip
        self.hpo_port = hpo_port
        self.hpo_status = hpo_status
        self.project = project
        self.algorithm = algorithm
        self.dataset = dataset
        self.dataset_hyperparam_name = dataset_hyperparam_name
        self.tasks = tasks
        self.err = err if isinstance(err, list) else []
        if load_detail:
            if not self.project or self.project_id != self.project.id:
                self.project = Project(id=self.project_id,
                                       load_detail=True)
            if not self.algorithm or self.algorithm_id != self.algorithm.id:
                self.algorithm = Algorithm(id=self.algorithm_id,
                                           load_detail=True)
            if not self.dataset or self.dataset_id != self.dataset.id:
                self.dataset = Dataset(id=self.dataset_id,
                                       load_detail=True)

    @classmethod
    def get_list(cls):
        """
        Listing is currently not supported for HpoExperiment
        """
        raise AnyLearnException("Listing is currently not supported for HpoExperiment")

    def from_dict(self, data: dict, load_detail=False):
        data = dict(
            {
                'project_id': self.project_id,
                'algorithm_id': self.algorithm_id,
                'dataset_id': self.dataset_id,
                'hpo_search_space': self.hpo_search_space,
                'hpo_max_runs': self.hpo_max_runs,
                'hpo_max_duration': self.hpo_max_duration,
                'hpo_tuner_name': self.hpo_tuner_name,
                'hpo_mode': self.hpo_mode,
                'hpo_concurrency': self.hpo_concurrency,
                'resource_request': self.resource_request,
                'created_at': self.created_at,
                'hpo_id': self.hpo_id,
                'hpo_ip': self.hpo_ip,
                'hpo_port': self.hpo_port,
                'hpo_status': self.hpo_status,
                'project': self.project,
                'algorithm': self.algorithm,
                'dataset': self.dataset,
                'tasks': self.tasks,
                'err': self.err,
            },
            **data,
        )
        if 'dataset_ids' in data and data['dataset_ids'] and isinstance(data['dataset_ids'], dict):
            data['dataset_id'] = list(data['dataset_ids'].values())[0].replace('$', '')
        self.__init__(
            project_id=data['project_id'],
            algorithm_id=data['algorithm_id'],
            # dataset_id=list(json.loads(data['dataset_ids']).values())[0].replace('$', ''),
            dataset_id=data['dataset_id'],
            hpo_search_space=(
                json.loads(data['hpo_search_space'])
                if isinstance(data['hpo_search_space'], str)
                else data['hpo_search_space']
            ),
            hpo_max_runs=data['hpo_max_runs'],
            hpo_max_duration=data['hpo_max_duration'],
            hpo_tuner_name=data['hpo_tuner_name'],
            hpo_mode=data['hpo_mode'],
            hpo_concurrency=data['hpo_concurrency'],
            resource_request=data['resource_request'],
            created_at=data['created_at'],
            hpo_id=data['hpo_id'],
            hpo_ip=data['hpo_ip'],
            hpo_port=data['hpo_port'],
            hpo_status=data['hpo_status'],
            tasks=(
                json.loads(data['tasks'])
                if isinstance(data['tasks'], str)
                else data['tasks']
            ),
            err=data['err'],
            load_detail=load_detail,
        )

    def get_detail(self):
        """
        以调参实验ID获取该实验的描述。

        - 对象属性 :obj:`id` 应为非空

        Returns
        -------
        HpoExperimentNniBack
            符合传入ID的调参实验描述对象。
        """

        data = get_with_token(f"{url_base()}/hpo",
                              params={'project_id': self.project_id})
        self.from_dict(data=data, load_detail=True)

    def stop(self):
        """
        停止调参实验。

        - 对象属性 :obj:`project_id` 应为非空
        """
        self._check_fields(required=['project_id'])
        logger.info(f"Trying to stop HPO of project {self.project_id}")
        res = post_with_token(f"{url_base()}/hpo/stop",
                              data={'project_id': self.project_id})
        return res

    def resume(self):
        """
        重启调参实验。

        - 对象属性 :obj:`project_id` 应为非空
        """
        self._check_fields(required=['project_id'])
        logger.info(f"Trying to resume HPO of project {self.project_id}")
        res = post_with_token(f"{url_base()}/hpo/resume",
                              data={'project_id': self.project_id})
        return res

    def view(self):
        """
        启动调参可视化面板。

        - 对象属性 :obj:`project_id` 应为非空
        """
        self._check_fields(required=['project_id'])
        logger.info(f"Trying to start HPO board of project {self.project_id}")
        res = post_with_token(f"{url_base()}/hpo/view",
                              data={'project_id': self.project_id})
        if not res or 'node_port' not in res:
            logger.error(f"Response error: {res}")
            raise AnyLearnException("请求未能得到有效响应")
        old = urlparse(AnylearnConfig.cluster_address)
        new = ParseResult(
            scheme=old.scheme,
            netloc=f"{old.hostname}:{res['node_port']}",
            path=old.path,
            params=old.params,
            query=old.query,
            fragment=old.fragment,
        )
        return {'url': new.geturl()}

    def get_tasks(self):
        self._check_fields(required=['project_id'])
        logger.info(f"Trying to fetch HPO tasks project {self.project_id}")
        res = get_with_token(f"{url_base()}/hpo/tasks",
                             params={'project_id': self.project_id})
        if not isinstance(res, dict):
            logger.error(f"Response error: {res}")
            raise AnyLearnException("请求未能得到有效响应")
        self.tasks = res
        return self.tasks

    def get_log(self):
        """
        获取调参实验日志（nnimanager.log）。

        - 对象属性 :obj:`project_id` 应为非空
        """
        self._check_fields(required=['project_id'])
        logger.info(f"Trying to retrieve HPO log of project {self.project_id}")
        res = get_with_token(f"{url_base()}/hpo/log",
                             params={'project_id': self.project_id})
        if not res or 'data' not in res:
            logger.error(f"Response error: {res}")
            raise AnyLearnException("请求未能得到有效响应")
        return res['data']

    def get_trial_logs(self):
        """
        获取调参实验日志（nnimanager.log）。

        - 对象属性 :obj:`project_id` 应为非空
        """
        self._check_fields(required=['project_id'])
        logger.info(f"Trying to retrieve HPO log of project {self.project_id}")
        logger.warning("Note that this method may be time-consuming")
        self.get_detail()
        self.get_tasks()
        if not self.tasks:
            return {}
        logs = {}
        for k in self.tasks.keys():
            res = get_with_token(f"{url_base()}/hpo/log/trial",
                                 params={'train_task_id': k})
            if not res or 'data' not in res:
                logger.warning(f"请求未能得到有效响应: {res}")
                continue
            logs[k] = res['data']
        return logs

    def get_trial_train_tasks(self) -> list[TrainTask]:
        self._check_fields(required=['project_id'])
        logger.info(f"Trying to retrieve HPO log of project {self.project_id}")
        self.get_detail()
        self.get_tasks()
        return [TrainTask(id=k, load_detail=True) for k in self.tasks.keys()] if self.tasks else []

    def get_best_train_task(self) -> TrainTask:
        """
        获取调参实验中取得最佳指标的训练任务。

        - 对象属性 :obj:`project_id` 应为非空
        """
        res = get_with_token(f"{url_base()}/hpo/best",
                             params={'project_id': self.project_id})
        if not res or not 'id' in res:
            logger.error(f"Response error: {res}")
            raise AnyLearnException("请求未能得到有效响应")
        task = TrainTask(id=res['id'], load_detail=True)
        task.get_final_metric()
        return task

    def export_best_model(self,
                          local_save_path: str,
                          downloader: Optional[ResourceDownloader]=None):
        """
        导出调参实验中取得最佳指标的模型至本地（下载）。

        - 对象属性 :obj:`project_id` 应为非空

        Parameters
        ----------
        local_save_path : :obj:`str`
            模型下载的本地保存路径。
        downloader : :obj:`ResourceDownloader`, optional
            模型下载器实现对象（建议留空）。
            缺省时将实例化一个内置的异步下载器。
        """
        best_train_task = self.get_best_train_task()
        if not downloader:
            downloader = AsyncResourceDownloader()
        t = Thread(target=Resource.download_file,
                   kwargs={
                       'resource_id': best_train_task.results_id,
                       'save_path': local_save_path,
                       'downloader': downloader, })
        logger.info("Downloading HPO best model...")
        t.start()
        t.join()
        logger.info("Successfully downloaded model.")

    def transform_best_model(
        self,
        model_name: str,
        model_description: Optional[str]=None,
        model_transform_polling: Union[float, int]=5
    ) -> Model:
        """
        转存调参实验中取得最佳指标的模型至Anylearn后端引擎。

        .. Note:: 训练的输出文件夹将被完整转存。

        Parameters
        ----------
        model_name : :obj:`str`
            模型名称。
        model_description : :obj:`str`, optional
            模型描述。
            默认为空。
        model_transform_polling : :obj:`float|int`, optional
            模型转存过程中，轮询模型转存状态的时间间隔。
            默认为5秒。

        Returns
        -------
        Model
            最佳模型转存后的对象。
        """
        best_train_task = self.get_best_train_task()
        model = best_train_task.transform_model(
            file_path=".",
            name=model_name,
            description=model_description,
            polling=model_transform_polling,
        )

        logger.info(
            "Transforming HPO best training "
            "result into Anylearn model..."
        )
        return model

    @classmethod
    def load(cls, data: dict, load_detail=False):
        """
        从接口返回值映射对一个调参实验描述进行实例化。

        Parameters
        ----------
        data : :obj:`dict`
            以字典承载的调参实验元信息。

        Returns
        -------
        HpoExperiment
            调参实验描述对象。
        """
        hpo = HpoExperiment()
        hpo.from_dict(data=data, load_detail=load_detail)
        return hpo

    def save(self):
        # Only create
        self._check_fields(required=self._fields['required']['create'])
        return self._create()

    def _update(self):
        raise AnyLearnException("Update is currently not supported for HpoExperiment")

    def _create(self):
        """
        创建对象，如果子类创建方法与此有较大差异可以重写此方法
        """
        data = self._payload_create()
        data['hpo_search_space'] = json.dumps(data['hpo_search_space'])
        data['dataset_ids'] = json.dumps({
            data['dataset_hyperparam_name']: f"${data['dataset_id']}"
        })
        data['resource_request'] = json.dumps(data['resource_request'])
        res = post_with_token(self._url_create(), data=data)
        if not res or 'project_id' not in res:
            logger.error(f"Response error: {res}")
            raise AnyLearnException("请求未能得到有效响应")
        return True

    def delete(self):
        raise AnyLearnException("Deletion is currently not supported for HpoExperiment")

    def _namespace(self):
        return "hpo"

    def _url_create(self):
        """
        创建对象url，如果子类创建对象接口名称不是 :obj:`add` ，可以重写此方法来定制接口名称
        """
        return f"{url_base()}/{self._namespace()}/start"

    def __repr__(self) -> str:
        kv = (
            f"    project_id = '{self.project_id}',\n"
            f"    algorithm_id = '{self.algorithm_id}',\n"
            f"    dataset_id = '{self.dataset_id}',\n"
            f"    hpo_search_space = '{self.hpo_search_space}',\n"
            f"    hpo_max_runs = '{self.hpo_max_runs}',\n"
            f"    hpo_max_duration = '{self.hpo_max_duration}',\n"
            f"    hpo_tuner_name = '{self.hpo_tuner_name}',\n"
            f"    hpo_mode = '{self.hpo_mode}',\n"
            f"    hpo_concurrency = '{self.hpo_concurrency}',\n"
            f"    created_at = '{self.created_at}',\n"
            f"    hpo_id = '{self.hpo_id}',\n"
            f"    hpo_ip = '{self.hpo_ip}',\n"
            f"    hpo_port = '{self.hpo_port}',\n"
            f"    hpo_status = '{self.hpo_status}',\n"
            f"    tasks = '{self.tasks}',\n"
            f"    err = '{self.err}',\n"
        )
        return f"HpoExperiment(\n{kv})"



from __future__ import annotations
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
import traceback
from typing import Optional, TYPE_CHECKING, Union

import sqlalchemy

from anylearn.config import AnylearnConfig
from anylearn.storage.db.models import (
    Base,
    SqlLocalAlgorithm,
    SqlLocalDataset,
    SqlLocalHpoExperiment,
    SqlLocalModel,
    SqlLocalTrainTask,
    SqlTrainProfile,
)
from anylearn.utils import logger
from anylearn.utils.errors import AnyLearnException
from anylearn.utils.singleton import Singleton

if TYPE_CHECKING:
    # Avoid circular import for type hinting (ENFER)
    from anylearn.applications.hpo_experiment import HpoExperiment
    from anylearn.interfaces import TrainTask


def _get_session(SessionMaker, db_type):
    """
    Creates a factory for producing exception-safe SQLAlchemy sessions that are made available
    using a context manager. Any session produced by this factory is automatically committed
    if no exceptions are encountered within its associated context. If an exception is
    encountered, the session is rolled back. Finally, any session produced by this factory is
    automatically closed when the session's associated context is exited.
    Credit: mlflow@/mlflow/store/db/utils.py
    """

    @contextmanager
    def make_managed_session():
        """Provide a transactional scope around a series of operations."""
        session = SessionMaker()
        try:
            if db_type == "sqlite":
                session.execute("PRAGMA foreign_keys = ON;")
                session.execute("PRAGMA case_sensitive_like = true;")
            yield session
            session.commit()
        except Exception as e:
            session.rollback()
            raise
        finally:
            session.close()
    return make_managed_session


class DB(metaclass=Singleton):
    def __init__(self):
        AnylearnConfig.init_workspace()
        self.alembic_ini = (
            Path(__file__).parent /
            "migrations" /
            "alembic.ini"
        )

        # DB physical concerns
        self.engine = sqlalchemy.create_engine(AnylearnConfig.db_uri)
        self.__upgrade_tables()

        # ORM logical concerns
        Base.metadata.bind = self.engine
        # Instance <xxx> is not bound to a Session; attribute refresh operation cannot proceed
        SessionMaker = sqlalchemy.orm.sessionmaker(bind=self.engine, expire_on_commit=False)
        self.session = _get_session(SessionMaker, "sqlite")

    def __upgrade_tables(self):
        from alembic import command

        logger.info("Upgrading database tables...")
        config = self.__get_alembic_config()
        with self.engine.begin() as connection:
            config.attributes["connection"] = connection # type: ignore
            config.attributes['configure_logger'] = False # type: ignore
            command.upgrade(config, "heads")

    def __get_alembic_config(self):
        from alembic.config import Config
        config = Config(self.alembic_ini)
        config.set_main_option('script_location', str(self.alembic_ini.parent))
        config.set_main_option('sqlalchemy.url', AnylearnConfig.db_uri)
        return config

    def force_init(self):
        self.__init__()

    def _save_to_db(self, session, objs):
        if type(objs) is list:
            session.add_all(objs)
        else:
            session.add(objs)

    def create_or_update_train_task(self,
                                    train_task: TrainTask,
                                    hpo_id: Optional[str] = None):
        with self.session() as s:
            train_task_obj = s.get(SqlLocalTrainTask, train_task.id)
            if train_task_obj:
                self.update_train_task(train_task=train_task)
            else:
                self.create_train_task(train_task=train_task, hpo_id=hpo_id)

    def create_train_task(self,
                          train_task: TrainTask,
                          hpo_id: Optional[str]=None):
        with self.session() as s:
            s.add(
                SqlLocalTrainTask(
                    id=train_task.id,
                    remote_state_sofar=train_task.state,
                    secret_key=train_task.secret_key,
                    project_id=train_task.project_id,
                    hpo_id=hpo_id,
                    hpo_final_metric=train_task.final_metric,
                )
            )
            try:
                s.commit()
            except Exception as e:
                traceback.format_exc()
                traceback.print_exc()
                logger.error(e)
                s.rollback()
                raise

    def delete_train_task(self, id: str):
        with self.session() as s:
            s.query(SqlLocalTrainTask)\
                .filter(SqlLocalTrainTask.id == id)\
                .delete()

    def delete_train_task_all(self):
        with self.session() as s:
            s.query(SqlLocalTrainTask).delete()

    def update_train_task(self, train_task: TrainTask):
        with self.session() as s:
            local_train_task = s.query(SqlLocalTrainTask)\
                .filter(SqlLocalTrainTask.id == train_task.id)\
                .first()
            if local_train_task:
                try:
                    local_train_task.id = train_task.id
                    local_train_task.remote_state_sofar = train_task.state
                    local_train_task.secret_key = train_task.secret_key
                    local_train_task.project_id = train_task.project_id
                    local_train_task.hpo_final_metric = train_task.final_metric
                except Exception as e:
                    traceback.format_exc()
                    traceback.print_exc()
                    logger.error(e)
                    s.rollback()
                    raise

    def get_unfinished_train_tasks(self):
        with self.session() as s:
            obj_list = s.query(SqlLocalTrainTask)\
                .filter(SqlLocalTrainTask.remote_state_sofar.in_([
                    # TODO: this is a work-around, make this generic
                    0,1
                ]))\
                .all()
            return obj_list

    def get_all_train_tasks(self):
        with self.session() as s:
            obj_list = s.query(SqlLocalTrainTask).all()
            return obj_list

    def get_train_task(self, id: str):
        with self.session() as s:
            obj = s.get(SqlLocalTrainTask, id)
            return obj
    
    def _check_resource_class(self, resource_class: type):
        if resource_class not in [
            SqlLocalAlgorithm,
            SqlLocalDataset,
            SqlLocalModel,
        ]:
            raise AnyLearnException((
                "'resource_class' must be one of "
                "["
                "SqlLocalAlgorithm,"
                "SqlLocalDataset,"
                "SqlLocalModel,"
                "]"
            ))

    def create_local_algorithm(self, id: Optional[str]=None, checksum: Optional[str]=None):
        return self._create_resource(
            resource_class=SqlLocalAlgorithm,
            resource_id=id,
            resource_checksum=checksum
        )

    def create_local_dataset(self, id: Optional[str]=None, checksum: Optional[str]=None):
        return self._create_resource(
            resource_class=SqlLocalDataset,
            resource_id=id,
            resource_checksum=checksum
        )

    def create_local_model(self, id: Optional[str]=None, checksum: Optional[str]=None):
        return self._create_resource(
            resource_class=SqlLocalModel,
            resource_id=id,
            resource_checksum=checksum
        )

    def _create_resource(self, resource_class: type, resource_id: Optional[str]=None,
                        resource_checksum: Optional[str]=None):
        self._check_resource_class(resource_class)
        with self.session() as s:
            s.add(
                resource_class(
                    id=resource_id,
                    checksum=resource_checksum,
                )
            )
            try:
                s.commit()
            except Exception as e:
                traceback.format_exc()
                traceback.print_exc()
                logger.error(e)
                s.rollback()
                raise

    def find_local_algorithm_by_checksum(self, checksum: str):
        return self._find_resource_by_checksum(
            resource_class=SqlLocalAlgorithm,
            checksum=checksum
        )

    def find_local_dataset_by_checksum(self, checksum: str):
        return self._find_resource_by_checksum(
            resource_class=SqlLocalDataset,
            checksum=checksum
        )

    def find_local_model_by_checksum(self, checksum: str):
        return self._find_resource_by_checksum(
            resource_class=SqlLocalModel,
            checksum=checksum
        )

    def _find_resource_by_checksum(self,
                                   resource_class: Union[
                                       SqlLocalAlgorithm,
                                       SqlLocalDataset,
                                       SqlLocalModel
                                   ],
                                   checksum: str):
        self._check_resource_class(resource_class)
        with self.session() as s:
            res = s.query(resource_class)\
                .filter(resource_class.checksum == checksum)\
                .first()
            return res.id if res else None

    def delete_local_algorithm(self, id: str):
        return self._delete_resource(
            resource_class=SqlLocalAlgorithm,
            id=id
        )

    def delete_local_dataset(self, id: str):
        return self._delete_resource(
            resource_class=SqlLocalDataset,
            id=id
        )

    def delete_local_model(self, id: str):
        return self._delete_resource(
            resource_class=SqlLocalModel,
            id=id
        )

    def _delete_resource(self, resource_class: type, id: str):
        self._check_resource_class(resource_class)
        with self.session() as s:
            res = s.query(resource_class)\
                .filter_by(id=id)\
                .delete()
            return res == 1

    def get_hpo_best_train_task(self,
                                project_id: str,
                                hpo_mode: str) -> TrainTask:
        orderby = (
            sqlalchemy.desc(SqlLocalTrainTask.hpo_final_metric)
            if hpo_mode == "maximize"
            else sqlalchemy.asc(SqlLocalTrainTask.hpo_final_metric)
        )
        with self.session() as s:
            res = s.query(SqlLocalTrainTask)\
                .filter_by(project_id=project_id)\
                .order_by(orderby)\
                .first()
            return TrainTask.from_sql(sql_local_train_task=res)

    def get_hpo_experiments(self) -> list[SqlLocalHpoExperiment]:
        with self.session() as s:
            return s.query(SqlLocalHpoExperiment).all()

    def get_hpo_experiment(self, id: str) -> SqlLocalHpoExperiment:
        with self.session() as s:
            return s.get(SqlLocalHpoExperiment, id)

    def create_hpo_experiment(self,
                              id: Optional[str]=None,
                              port: Optional[int]=None,
                              mode: Optional[str]=None,
                              status: Optional[str]=None,
                              project_id: Optional[str]=None,
                              project_name: Optional[str]=None,
                              algorithm_id: Optional[str]=None,
                              dataset_id: Optional[str]=None,
                              algorithm_dir: Optional[str]=None,
                              algorithm_archive: Optional[str]=None,
                              dataset_dir: Optional[str]=None,
                              dataset_archive: Optional[str]=None,
                              created_at: Optional[datetime]=None):
        with self.session() as s:
            exp = SqlLocalHpoExperiment(
                id=id,
                port=port,
                mode=mode,
                status=status,
                project_id=project_id,
                project_name=project_name,
                algorithm_id=algorithm_id,
                algorithm_dir=algorithm_dir,
                algorithm_archive=algorithm_archive,
                dataset_id=dataset_id,
                dataset_dir=dataset_dir,
                dataset_archive=dataset_archive,
                created_at=created_at,
            )
            self._save_to_db(objs=exp, session=s)

    def update_hpo_experiment_status(self, id: str, status: str):
        with self.session() as s:
            exp = s.get(SqlLocalHpoExperiment, id)
            if not exp:
                raise AnyLearnException(f"Local HPO experiment {id} does not exist")
            exp.status = status
            self._save_to_db(objs=exp, session=s)

    def update_hpo_experiment_port(self, id: str, port: int):
        with self.session() as s:
            exp = s.get(SqlLocalHpoExperiment, id)
            if not exp:
                raise AnyLearnException(f"Local HPO experiment {id} does not exist")
            exp.port = port
            self._save_to_db(objs=exp, session=s)

    def get_train_profile(self, id: Optional[str] = None) -> SqlTrainProfile:
        with self.session() as s:
            return s.get(SqlTrainProfile, id)

    def create_train_profile(self,
                             id: str,
                             train_task_id: str,
                             algorithm_id: Optional[str] = None,
                             algorithm_dir: Optional[str] = None,
                             algorithm_archive: Optional[str] = None,
                             dataset_id: Optional[str] = None,
                             dataset_dir: Optional[str] = None,
                             dataset_archive: Optional[str] = None,
                             entrypoint: Optional[str] = None,
                             train_params: Optional[str] = None,):
        with self.session() as s:
            exp = SqlTrainProfile(
                id=id,
                train_task_id=train_task_id,
                entrypoint=entrypoint,
                train_params=train_params,
                algorithm_id=algorithm_id,
                algorithm_dir=algorithm_dir,
                algorithm_archive=algorithm_archive,
                dataset_id=dataset_id,
                dataset_dir=dataset_dir,
                dataset_archive=dataset_archive,
            )
            self._save_to_db(objs=exp, session=s)



import responses
from requests.exceptions import RequestException

from anylearn.config import init_sdk
from anylearn.interfaces.resource.resource import ResourceState
from anylearn.interfaces.resource.algorithm import Algorithm
from anylearn.utils.errors import AnyLearnException, AnyLearnMissingParamException
from tests.base_test_case import BaseTestCase

class TestAlgorithm(BaseTestCase):
    @responses.activate
    def test_list_algo_ok(self):
        responses.add(responses.GET, url=self._url("algorithm/list"),
                      json=[
                          {
                              'id': "ALGO001",
                              'name': "TestAlgo1",
                              'description': "test",
                              'state': ResourceState.READY,
                              'public': False,
                              'upload_time': "2020-01-01 00:00",
                              'tags': "tag1, tag2",
                              'follows_anylearn_norm': True,
                              'git_address': "http://anylearn.testing/anylearn/TestAlgo1.git",
                              'git_migrated': True,
                          },
                          {
                              'id': "ALGO002",
                              'name': "TestAlgo2",
                              'description': "test",
                              'state': ResourceState.READY,
                              'public': True,
                              'upload_time': "2020-01-02 00:00",
                              'tags': "tag3, tag4",
                              'follows_anylearn_norm': False,
                              'git_address': "http://anylearn.testing/anylearn/TestAlgo2.git",
                              'git_migrated': True,
                          },
                      ],
                      status=200)
        algos = Algorithm.get_list()
        self.assertIsInstance(algos, list)
        self.assertEqual(len(algos), 2)
        self.assertIsInstance(algos[0], Algorithm)
        self.assertIsInstance(algos[1], Algorithm)
        self.assertEqual(algos[0].id, "ALGO001")
        self.assertEqual(algos[1].id, "ALGO002")
        self.assertEqual(algos[0].tags, "tag1, tag2")
        self.assertEqual(algos[1].tags, "tag3, tag4")
        self.assertEqual(algos[0].follows_anylearn_norm, True)
        self.assertEqual(algos[1].follows_anylearn_norm, False)

    @responses.activate
    def test_list_algo_ko_unknown_response(self):
        responses.add(responses.GET, url=self._url("algorithm/list"),
                      json={'msg': "Unknown response"},
                      status=200)
        with self.assertRaises(AnyLearnException) as ctx:
            Algorithm.get_list()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_get_algo_detail_explicit_call_ok(self):
        responses.add(responses.GET, url=self._url("algorithm/query?id=ALGO001"),
                      match_querystring=True, json=[{
                          'id': "ALGO001",
                          'name': "TestAlgo1",
                          'description': "test",
                          'state': ResourceState.READY,
                          'public': True,
                          'upload_time': "2020-01-01 00:00",
                          'filename': "test.tar.gz",
                          'is_zipfile': 1,
                          'file': "USER001/algos/ALGO001/test.tar.gz",
                          'size': "250.41",
                          'creator_id': "USER001",
                          'node_id': "NODE001",
                          'tags': "tag1, tag2",
                          'train_params': '[{"test": "test"}]',
                          'mirror_id': "MIRR001",
                          'follows_anylearn_norm': True,
                          'git_address': "http://anylearn.testing/user001/TestAlgo1.git",
                          'git_migrated': True,
                      }],
                      status=200)
        algo = Algorithm(id="ALGO001")
        algo.get_detail()
        self.assertEqual(algo.filename, "test.tar.gz")
        self.assertTrue(algo.is_zipfile)
        self.assertEqual(algo.file_path, "USER001/algos/ALGO001/test.tar.gz")
        self.assertEqual(algo.size, "250.41")
        self.assertEqual(algo.creator_id, "USER001")
        self.assertEqual(algo.node_id, "NODE001")
        self.assertEqual(algo.tags, "tag1, tag2")
        self.assertIsInstance(algo.train_params, list)
        self.assertEqual(str(algo.train_params), "[{'test': 'test'}]")
        self.assertEqual(algo.mirror_id, "MIRR001")
        self.assertEqual(algo.follows_anylearn_norm, True)

    @responses.activate
    def test_get_algo_detail_no_explicit_call_ok(self):
        responses.add(responses.GET, url=self._url("algorithm/query?id=ALGO001"),
                      match_querystring=True, json=[{
                          'id': "ALGO001",
                          'name': "TestAlgo1",
                          'description': "test",
                          'state': ResourceState.READY,
                          'public': True,
                          'upload_time': "2020-01-01 00:00",
                          'filename': "test.tar.gz",
                          'is_zipfile': 1,
                          'file': "USER001/algos/ALGO001/test.tar.gz",
                          'size': "250.41",
                          'creator_id': "USER001",
                          'node_id': "NODE001",
                          'tags': "tag1, tag2",
                          'train_params': '[{"test": "test"}]',
                          'mirror_id': "MIRR001",
                          'follows_anylearn_norm': False,
                          'git_address': "http://anylearn.testing/user001/TestAlgo1.git",
                          'git_migrated': True,
                      }],
                      status=200)
        algo = Algorithm(id="ALGO001", load_detail=True)
        self.assertEqual(algo.name, "TestAlgo1")
        self.assertEqual(algo.filename, "test.tar.gz")
        self.assertTrue(algo.is_zipfile)
        self.assertEqual(algo.file_path, "USER001/algos/ALGO001/test.tar.gz")
        self.assertEqual(algo.size, "250.41")
        self.assertEqual(algo.creator_id, "USER001")
        self.assertEqual(algo.node_id, "NODE001")
        self.assertEqual(algo.tags, "tag1, tag2")
        self.assertIsInstance(algo.train_params, list)
        self.assertEqual(str(algo.train_params), "[{'test': 'test'}]")
        self.assertEqual(algo.mirror_id, "MIRR001")
        self.assertEqual(algo.follows_anylearn_norm, False)

    @responses.activate
    def test_get_algo_detail_ko_403(self):
        responses.add(responses.GET, url=self._url("algorithm/query?id=ALGO403"),
                      match_querystring=True, status=403)
        algo = Algorithm(id="ALGO403")
        with self.assertRaises(RequestException) as ctx:
            algo.get_detail()
        e = ctx.exception
        self.assertIsInstance(e, RequestException)
        self.assertEqual(e.response.status_code, 403)

    @responses.activate
    def test_get_algo_detail_ko_unknown_response(self):
        responses.add(responses.GET, url=self._url("algorithm/query?id=ALGO250"),
                      match_querystring=True, json={'msg': "Unknown response"},
                      status=200)
        algo = Algorithm(id="ALGO250")
        with self.assertRaises(AnyLearnException) as ctx:
            algo.get_detail()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_get_algo_detail_empty_list(self):
        responses.add(responses.GET, url=self._url("algorithm/query?id=ALGO250"),
                      match_querystring=True, json=[],
                      status=200)
        algo = Algorithm(id="ALGO250")
        with self.assertRaises(AnyLearnException) as ctx:
            algo.get_detail()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    def test_create_algo_name_not_standard(self):
        with self.assertRaises(AnyLearnException) as ctx:
            algo = Algorithm(name="Test@ALGO/001", 
                             filename="test.tar.gz",
                             mirror_id="MIRR001")
            algo.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(
            e.msg,
            "Algorithm name should contain only alphanumeric, dash ('-'), underscore ('_') and dot ('.') characters"
        )

    @responses.activate
    def test_create_algo_ok(self):
        responses.add(responses.POST, url=self._url("algorithm/add"),
                      json={'data': "ALGO001", 'message': "算法添加成功"},
                      status=200)
        algo = Algorithm(name="Test-ALGO_00.1", filename="test.tar.gz",
                         train_params='[{"name": "test", "type": "int", "suggest": 1, "default": "0", "scope": "<0,1>"}]',
                         mirror_id="MIRR001")
        result = algo.save()
        self.assertTrue(result)
        self.assertEqual(algo.id, "ALGO001")

    def test_create_algo_ko_empty_name_filename_mirror_params(self):
        algo = Algorithm()
        with self.assertRaises(AnyLearnMissingParamException) as ctx:
            algo.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnMissingParamException)
        self.assertEqual(e.msg, "Algorithm缺少必要字段：['name', 'mirror_id']")

    # def test_create_algo_ko_empty_entrypoint_training_output_training(self):
    #     algo = Algorithm(name="TestALGO250", filename="test.tar.gz",
    #                      train_params='[{"name": "test", "type": "int", "suggest": 1, "default": "0", "scope": "<0,1>"}]',
    #                      mirror_id="MIRR001", follows_anylearn_norm=False)
    #     with self.assertRaises(AnyLearnMissingParamException) as ctx:
    #         algo.save()
    #     e = ctx.exception
    #     self.assertIsInstance(e, AnyLearnMissingParamException)
    #     self.assertEqual(e.msg, "Algorithm缺少必要字段：['entrypoint_training', 'output_training']")

    @responses.activate
    def test_create_algo_ko_unknown_response(self):
        responses.add(responses.POST, url=self._url("algorithm/add"),
                      json={'msg': "Unknown response"}, status=201)
        algo = Algorithm(name="TestAlgo", filename="ALGO250.tar.gz",
                         train_params='[{"test": "test"}]',
                         mirror_id="MIRR001")
        with self.assertRaises(AnyLearnException) as ctx:
            algo.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_update_algo_ok(self):
        responses.add(responses.PUT, url=self._url("algorithm/update"),
                      json={'data': "ALGO001", 'message': "算法信息更新成功"},
                      status=200)
        algo = Algorithm(id="ALGO001", name="TestAlgo1", description="test m",
                          state=ResourceState.READY,
                          public=True,
                          upload_time="2020-01-01 00:00")
        result = algo.save()
        self.assertTrue(result)

    def test_update_algo_name_not_standard(self):
        with self.assertRaises(AnyLearnException) as ctx:
            algo = Algorithm(id="ALGO123",
                             name="Test@ALGO/001",
                             filename="test.tar.gz",
                             mirror_id="MIRR001")
            algo.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(
            e.msg,
            "Algorithm name should contain only alphanumeric, dash ('-'), underscore ('_') and dot ('.') characters"
        )

    @responses.activate
    def test_update_algo_ko_empty_name(self):
        algo = Algorithm(id="ALGO250", name="")
        with self.assertRaises(AnyLearnMissingParamException) as ctx:
            algo.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnMissingParamException)
        self.assertEqual(e.msg, "Algorithm缺少必要字段：['name']")

    @responses.activate
    def test_update_algo_ko_unknown_response(self):
        responses.add(responses.PUT, url=self._url("algorithm/update"),
                      json={'msg': "Unknown response"},
                      status=200)
        algo = Algorithm(id="ALGO250", name="TestAlgo", description="test",
                          state=ResourceState.READY,
                          public=False,
                          upload_time="2020-01-01 02:50")
        with self.assertRaises(AnyLearnException) as ctx:
            algo.save()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    @responses.activate
    def test_delete_algo_ok(self):
        responses.add(responses.DELETE,
                      url=self._url("algorithm/delete?id=ALGO001&force=0"),
                      match_querystring=True,
                      json={'data': "ALGO001", 'message': "算法删除成功"},
                      status=200)
        algo = Algorithm(id="ALGO001", name="TestAlgo1", description="test d",
                          state=ResourceState.READY,
                          public=True,
                          upload_time="2020-01-01 00:00")
        result = algo.delete()
        self.assertTrue(result)

    @responses.activate
    def test_delete_algo_ko_unknown_response(self):
        responses.add(responses.DELETE,
                      url=self._url("algorithm/delete?id=ALGO250&force=0"),
                      match_querystring=True,
                      json={'msg': "Unknown response"},
                      status=204)
        algo = Algorithm(id="ALGO250", name="TestAlgo", description="test",
                          state=ResourceState.READY,
                          public=False,
                          upload_time="2020-01-01 02:50")
        with self.assertRaises(AnyLearnException) as ctx:
            algo.delete()
        e = ctx.exception
        self.assertIsInstance(e, AnyLearnException)
        self.assertEqual(e.msg, "请求未能得到有效响应")

    def test_algo_params_parsing(self):
        json_params = '''
            [
                {"name": "test0", "type": "int", "suggest": 1, "scope": "<-Infinity,Infinity>"},
                {"name": "test1", "type": "int", "suggest": 1, "default": "1", "scope": "<0,1>"},
                {"name": "test2", "type": "int", "suggest": 1, "default": "2", "scope": "<0,2>"}
            ]
        '''
        algo = Algorithm(train_params=json_params)
        
        self.assertIsInstance(algo.required_train_params, list)
        self.assertEqual(len(algo.required_train_params), 1)
        self.assertEqual(algo.required_train_params[0]['name'], 'test0')
        self.assertIsInstance(algo.default_train_params, dict)
        self.assertEqual(len(algo.default_train_params), 2)
        self.assertEqual(algo.default_train_params['test1'], '1')
        self.assertEqual(algo.default_train_params['test2'], '2')



from __future__ import annotations

from anylearn.utils.api import get_with_token, url_base
from anylearn.utils.errors import AnyLearnException


class Project:
    """
    AnyLearn训练项目类

    Attributes
    ----------
    id
        项目的唯一标识符，自动生成，由PROJ+uuid1生成的编码中后28个有效位（小写字母和数字）组成
    name
        项目名称（非空 长度1~50）
    description
        项目描述（可为空 长度最大255）
    created_at
        创建时间
    updated_at
        更新时间
    creator_id
        创建者的ID
    creator_username
    """

    def __init__(self, *_, **kwargs):
        self.id = kwargs.get('id', None)
        self.name = kwargs.get('name', None)
        self.description = kwargs.get('description', None)
        self.created_at = kwargs.get('create_time', None)
        self.updated_at = kwargs.get('update_time', None)
        self.creator_id = kwargs.get('creator_id', None)
        self.creator_username = kwargs.get('creator_username', None)

    @classmethod
    def from_full_name(cls, full_name: str) -> Project:
        res = get_with_token(
            f"{url_base()}/project/query",
            params={'fullname': full_name},
        )
        if not res or not isinstance(res, list):
            raise AnyLearnException("Request failed")
        return Project(**res[0])



from __future__ import annotations
from datetime import datetime
from typing import Optional

from anylearn.utils.errors import AnyLearnException
from anylearn.storage.db import DB
from anylearn.utils import utc_plus_8


class TrainProfile:
    """
    训练任务描述类。

    一个训练任务的描述包括了
    训练任务命令、
    训练参数、
    算法、
    数据集、
    训练的元信息
    等几方面，
    用户可以根据这些信息可以在多次训练之间进行区别和回忆。
    任务描述中封装了一些与训练相关的功能，
    如：获取本地训练的训练描述集合、创建训练描述、获取某一训练任务的训练描述等等，
    以便对任一训练任务的生命周期进行完整的回溯和管理。

    Attributes
    ----------
    id : :obj:`str`
        训练任务描述ID。
    train_task_id : :obj:`str`
        训练任务ID。
    entrypoint : :obj:`str`, optional
        启动训练的入口命令。
    train_params : :obj:`str`
        训练任务参数。
    algorithm_id : :obj:`str`
        已在Anylearn远程注册的算法ID。
    algorithm_dir : :obj:`str`, optional
        本地算法目录路径。
    algorithm_archive : :obj:`str`, optional
        本地算法压缩包路径。
    dataset_id : :obj:`str`
        已在Anylearn远程注册的数据集ID。
    dataset_dir : :obj:`str`, optional
        本地数据集目录路径。
    dataset_archive : :obj:`str`, optional
        本地数据集压缩包路径。
    created_at : :obj:`str`, optional
        调参实验的创建时间。
        默认为东八时区的当前时间。
    """

    def __init__(self,
                 id: str,
                 train_task_id: str,
                 entrypoint: Optional[str] = None,
                 algorithm_id: Optional[str] = None,
                 dataset_id: Optional[str] = None,
                 train_params: Optional[str] = None,
                 algorithm_dir: Optional[str] = None,
                 algorithm_archive: Optional[str] = None,
                 dataset_dir: Optional[str] = None,
                 dataset_archive: Optional[str] = None,
                 created_at: Optional[datetime] = utc_plus_8()):
        """
        Parameters
        ----------
        id : :obj:`str`
            训练任务描述ID。
        train_task_id : :obj:`str`
            训练任务ID。
        entrypoint : :obj:`str`, optional
            启动训练的入口命令。
        train_params : :obj:`str`
            训练任务参数。
        algorithm_id : :obj:`str`
            已在Anylearn远程注册的算法ID。
        algorithm_dir : :obj:`str`, optional
            本地算法目录路径。
        algorithm_archive : :obj:`str`, optional
            本地算法压缩包路径。
        dataset_id : :obj:`str`
            已在Anylearn远程注册的数据集ID。
        dataset_dir : :obj:`str`, optional
            本地数据集目录路径。
        dataset_archive : :obj:`str`, optional
            本地数据集压缩包路径。
        created_at : :obj:`str`, optional
            调参实验的创建时间。
            默认为东八时区的当前时间。
        """
        self.id = id
        self.train_task_id = train_task_id
        self.entrypoint = entrypoint
        self.train_params = train_params
        self.algorithm_id = algorithm_id
        self.dataset_id = dataset_id
        self.algorithm_dir = algorithm_dir
        self.algorithm_archive = algorithm_archive
        self.dataset_dir = dataset_dir
        self.dataset_archive = dataset_archive
        self.created_at = created_at

    @classmethod
    def get(cls, id: str=None, train_task_id: str=None):
        """
        以训练任务ID或描述ID获取该训练的描述。

        Parameters
        ----------
        id : :obj:`str`
            训练描述的ID。
        train_task_id : :obj:`str`
            训练任务的ID。

        Returns
        -------
        TrainProfile
            符合传入ID的训练描述对象。
        """
        if train_task_id:
            local_train_task = DB().get_train_task(id=train_task_id)
            if not local_train_task:
                raise AnyLearnException(f"资源{train_task_id}未找到")
            sql_train_profile = local_train_task.train_profile
        else:
            sql_train_profile = DB().get_train_profile(id=id)
        if not len(sql_train_profile):
            raise AnyLearnException(
                f"资源{train_task_id if train_task_id else id}未找到TrainProfile")
        return cls.from_sql(sql_train_profile[0])

    @classmethod
    def from_sql(cls, sql_train_profile):
        """
        从本地数据库模型映射对一个训练任务描述进行实例化。

        Parameters
        ----------
        sql_train_profile : :obj:`SqlTrainProfile`
            本地数据库映射对象。
        
        Returns
        -------
        TrainProfile
            符合传入映射的训练描述对象。
        """
        return TrainProfile(
            id=sql_train_profile.id,
            train_task_id=sql_train_profile.train_task_id,
            entrypoint=sql_train_profile.entrypoint,
            train_params=sql_train_profile.train_params,
            algorithm_id=sql_train_profile.algorithm_id,
            dataset_id=sql_train_profile.dataset_id,
            algorithm_dir=sql_train_profile.algorithm_dir,
            algorithm_archive=sql_train_profile.algorithm_archive,
            dataset_dir=sql_train_profile.dataset_dir,
            dataset_archive=sql_train_profile.dataset_archive,
            created_at=sql_train_profile.created_at,
        ) if sql_train_profile else None

    def create_in_db(self):
        """
        将训练任务描述写入本地数据库。
        """
        DB().create_train_profile(
            id=self.id,
            train_task_id=self.train_task_id,
            entrypoint=self.entrypoint,
            train_params=self.train_params,
            algorithm_id=self.algorithm_id,
            algorithm_dir=self.algorithm_dir,
            algorithm_archive=self.algorithm_archive,
            dataset_id=self.dataset_id,
            dataset_dir=self.dataset_dir,
            dataset_archive=self.dataset_archive,
        )

    @classmethod
    def get_train_profiles(cls, local_train_tasks=None):
        """
        以训练任务集合获取他们所有的训练描述。

        Parameters
        ----------
        local_train_tasks : :obj:`List[SqlLocalTrainTask]`
            本地训练任务集合。

        Returns
        -------
        List[TrainProfile]
        """
        return [cls.from_sql(local_train_task.train_profile[0] if len(local_train_task.train_profile) else None)
                for local_train_task in local_train_tasks]

    def __repr__(self) -> str:
        kv = (
            f"    id='{self.id}',\n"
            f"    train_task_id={self.train_task_id},\n"
            f"    entrypoint='{self.entrypoint}',\n"
            f"    train_params='{self.train_params}',\n"
            f"    algorithm_id='{self.algorithm_id}',\n"
            f"    algorithm_dir='{self.algorithm_dir}',\n"
            f"    algorithm_archive='{self.algorithm_archive}',\n"
            f"    dataset_id='{self.dataset_id}',\n"
            f"    dataset_dir='{self.dataset_dir}',\n"
            f"    dataset_archive='{self.dataset_archive}',\n"
            f"    created_at=datetime.strptime('{self.created_at.strftime('%Y-%m-%d %H:%M:%S')}', '%Y-%m-%d %H:%M:%S'),\n"
        )
        return f"TrainProfile(\n{kv})"



import io

import setuptools


CLIENT_VERSION = "0.17.1"
PACKAGE_NAME = "anylearn"

try:
    with io.open("README.md", encoding="utf-8") as f:
        LONG_DESCRIPTION = f.read()
except FileNotFoundError:
    LONG_DESCRIPTION = ""

REQUIRES = []
with open('requirements.txt') as f:
    for line in f:
        line, _, _ = line.partition('#')
        line = line.strip()
        REQUIRES.append(line)

setuptools.setup(
    name=PACKAGE_NAME,
    version=CLIENT_VERSION,
    license="Proprietary",
    description="Anylearn Python SDK",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    author_email="anylearn@nelbds.org.cn",
    author="Dmagine",
    install_requires=REQUIRES,
    python_requires='>=3.7',
    packages=setuptools.find_packages(),
    package_data={
        'anylearn': [
            "storage/db/migrations/alembic.ini",
        ],
    },
    include_package_data=True,
    classifiers=[
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    entry_points = '''
        [console_scripts]
        anyctl = anylearn.cli.cli:cli
    ''',
)
